# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect


# ------------------------------------------------------------------------------------
# this is the missing stuff when running python.exe compared with mayapy.exe

# mayaver = 2022
# pythonver = 37

# mayapath = '%ProgramFiles%/Autodesk/Maya{}'.format(mayaver)

# os.environ['MAYA_LOCATION'] = mayapath
# os.environ['PYTHONHOME'] = mayapath+'/Python{}'.format(mayaver, pythonver)
# os.environ['PATH'] = mayapath+'/bin;'.format(mayaver) + os.environ['PATH']

# from pythontk import File
# for d in [
# 	'{}/bin'.format(mayapath), 
# 	'{}/bin3'.format(mayapath), 
# 	'{}/Python{}'.format(mayapath, pythonver)
# 	]:
# 	for dd in File.getDirContents(d, 'dirpaths', excDirs='Python27',  recursive=True):
# 		print (dd)
# 		sys.path.append(dd)

# import maya.standalone
# maya.standalone.initialize(name='python')

import pymel.core as pm
# # ------------------------------------------------------------------------------------

from mayatk import *
import test


sfr = pm.melGlobals['cmdScrollFieldReporter']
pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)



class Edittk_test(Main, Edittk):
	'''
	'''
	#Tear down the any previous test by creating a new scene:
	pm.mel.file(new=True, force=True)

	#assemble the test scene:
	if not pm.objExists('cube1'):
		cube1 = pm.polyCube(width=5, height=5, depth=5, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1, name='cube1')

	if not pm.objExists('cube2'):
		cube2 = pm.polyCube(width=2, height=4, depth=8, subdivisionsX=3, subdivisionsY=3, subdivisionsZ=3, name='cube2')

	if not pm.objExists('cyl'):
		cyl = pm.polyCylinder(radius=5, height=10, subdivisionsX=12, subdivisionsY=1, subdivisionsZ=1, name='cyl')

	def test_rename(self):
		'''
		'''
		self.perform_test({
			"self.rename('cube1', 'newName')": None,
			"self.rename('newName', 'cube1')": None,
		})

	def test_setCase(self):
		'''
		'''
		self.perform_test({
			"self.setCase('cube1', 'lower')": None,
		})

	def test_setSuffixByObjLocation(self):
		'''
		'''
		if not pm.objExists('c1'):
			c1 = pm.polyCube(width=2, height=2, depth=8, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1, name='c1')

		if not pm.objExists('c2'):
			c2 = pm.polyCube(width=8, height=2, depth=2, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1, name='c2')
			pm.move(0, 0, 5, c2)

		self.perform_test({
			"self.setSuffixByObjLocation(['c1', 'c2'])": None,
		})

	def test_snapClosestVerts(self):
		'''
		'''
		self.perform_test({
			"self.snapClosestVerts('cube1', 'cube2')": None,
		})

	def test_mergeVertices(self):
		'''
		'''
		self.perform_test({
			"self.mergeVertices('cube1')": None,
		})

	def test_deleteAlongAxis(self):
		'''
		'''
		self.perform_test({
			"self.deleteAlongAxis('cube1')": None,
		})

	def test_getAllFacesOnAxis(self):
		'''
		'''
		self.perform_test({
			"self.getAllFacesOnAxis('cube1')": [], #faces should have been deleted by the previous test 'deleteAlongAxis'.
		})

	def test_cleanGeometry(self):
		'''
		'''
		self.perform_test({
			"self.cleanGeometry('cyl')": None,
		})

	def test_getOverlappingDupObjects(self):
		'''
		'''
		self.perform_test({
			"self.getOverlappingDupObjects(['cyl', 'cube1', 'cube2'])": set(),
		})

	def test_findNonManifoldVertex(self):
		'''
		'''
		self.perform_test({
			"self.findNonManifoldVertex('cyl')": set(),
		})

	def test_splitNonManifoldVertex(self):
		'''
		'''
		self.perform_test({
			"self.splitNonManifoldVertex('cyl')": None,
		})

	def test_getNGons(self):
		'''
		'''
		self.perform_test({
			"self.getNGons('cyl')": [],
		})

	def test_getOverlappingVertices(self):
		'''
		'''
		self.perform_test({
			"self.getOverlappingVertices('cyl')": [],
		})

	def test_getOverlappingFaces(self):
		'''
		'''
		self.perform_test({
			"self.getOverlappingFaces('cyl')": [],
			"self.getOverlappingFaces('cyl.f[:]')": [],
		})

	def test_getSimilarMesh(self):
		'''
		'''
		self.perform_test({
			"self.getSimilarMesh('cyl')": [],
		})

	def test_getSimilarTopo(self):
		'''
		'''
		self.perform_test({
			"self.getSimilarTopo('cyl')": [],
		})

# -----------------------------------------------------------------------------

if __name__=='__main__':

	unittest.main(exit=False)






# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})
# """

# # Deprecated ---------------------


# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect

import pymel.core as pm

from mayatk import *
import test


sfr = pm.melGlobals['cmdScrollFieldReporter']
pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)



class Comptk_test(Main, Comptk):
	'''
	set object mode:
		pm.selectMode(object=1)

	set component mode:
		pm.selectMode(component=1)

	set component mode type:
		pm.selectType(allObjects=1)
		pm.selectType(mc=1)
		pm.selectType(vertex=1)
		pm.selectType(edge=1)
		pm.selectType(facet=1)
		pm.selectType(polymeshUV=1)
		pm.selectType(meshUVShell=1)
	'''
	#Tear down the any previous test by creating a new scene:
	pm.mel.file(new=True, force=True)

	#assemble the test scene:
	if not pm.objExists('cyl'):
		cyl = pm.polyCylinder(radius=5, height=10, subdivisionsX=12, subdivisionsY=1, subdivisionsZ=1, name='cyl')

	if not pm.objExists('pln'):
		pln = pm.polyPlane(width=20, height=20, subdivisionsX=3, subdivisionsY=3, name='pln')

	if not pm.objExists('sph'):
		sph = pm.polySphere(radius=8, subdivisionsX=6, subdivisionsY=6, name='sph')

	def test_getComponentType(self):
		'''
		'''
		self.perform_test({
			"self.getComponentType('cyl.e[:]')": 'e',
			"self.getComponentType('cyl.vtx[:]', 'abv')": 'vtx',
			"self.getComponentType('cyl.e[:]', 'int')": 32,
			"self.getComponentType('cyl.e[:]', 'hex')": 0x8000,
		})

	def test_convertAlias(self):
		'''
		'''
		self.perform_test({
			"self.convertAlias('vertex', 'hex')": 0x0001,
			"self.convertAlias(0x0001, 'full')": 'Polygon Vertex',
		})

	def test_convertComponentType(self):
		'''
		'''
		self.perform_test({
			"self.convertComponentType('cylShape.vtx[:2]', 'vertex')": ['cylShape.vtx[0:2]'],
			"self.convertComponentType('cylShape.vtx[:2]', 'face')": ['cylShape.f[0:2]', 'cylShape.f[11:14]', 'cylShape.f[23]'],
			"self.convertComponentType('cylShape.vtx[:2]', 'edge')": ['cylShape.e[0:2]', 'cylShape.e[11]', 'cylShape.e[24:26]', 'cylShape.e[36:38]'],
			"self.convertComponentType('cylShape.vtx[:2]', 'uv')": ['cylShape.map[0:2]', 'cylShape.map[12:14]', 'cylShape.map[24]'],
		})

	def test_convertIntToComponent(self):
		'''
		'''
		self.perform_test({
			"self.convertIntToComponent('cyl', range(4), 'f')": ['cylShape.f[0:3]'],
			"self.convertIntToComponent('cyl', range(4), 'f', 'int', flatten=True)": [0, 1, 2, 3],
		})

	def test_filterComponents(self):
		'''
		'''
		self.perform_test({
			"self.filterComponents('cyl.vtx[:]', 'cyl.vtx[:2]', 'cyl.vtx[1:23]')": ['cylShape.vtx[0]'],
			"self.filterComponents('cyl.f[:]', range(2), range(1, 23))": ['cylShape.f[0]'],
		})

	def test_getComponents(self):
		'''
		'''
		self.perform_test({
			"self.getComponents('cyl', 'vertex', 'str', '', 'cyl.vtx[2:23]')": ['cylShape.vtx[0]', 'cylShape.vtx[1]', 'cylShape.vtx[24]', 'cylShape.vtx[25]'],
			"str(self.getComponents('cyl', 'vertex', 'cyl', '', 'cyl.vtx[:23]'))": "[MeshVertex('cylShape.vtx[24]'), MeshVertex('cylShape.vtx[25]')]",
			"self.getComponents('cyl', 'f', 'int')": [0, 35],
			"self.getComponents('cyl', 'edges')": ['cylShape.e[0:59]'],
			"self.getComponents('cyl', 'edges', 'str', 'cyl.e[:2]')": ['cylShape.e[0]', 'cylShape.e[1]', 'cylShape.e[2]'],
		})

	def test_getContigiousEdges(self):
		'''
		'''
		self.perform_test({
			"self.getContigiousEdges(['cyl.e[:2]'])": [{'cylShape.e[1]', 'cylShape.e[0]', 'cylShape.e[2]'}],
			"self.getContigiousEdges(['cyl.f[0]'])": [{'cylShape.e[24]', 'cylShape.e[0]', 'cylShape.e[25]', 'cylShape.e[12]'}],
		})

	def test_getContigiousIslands(self):
		'''
		'''
		self.perform_test({
			"self.getContigiousIslands('cyl.f[21:26]')": [{'cylShape.f[22]', 'cylShape.f[21]', 'cylShape.f[23]'}, {'cylShape.f[26]', 'cylShape.f[24]', 'cylShape.f[25]'}],
		})

	def test_getIslands(self):
		'''
		'''
		if not pm.objExists('cmb'): #create two objects and combine them.
			cmbA = pm.polyCylinder(radius=5, height=10, subdivisionsX=5, subdivisionsY=1, subdivisionsZ=1, name='cmbA')
			cmbB = pm.polyCylinder(radius=5, height=6, subdivisionsX=5, subdivisionsY=1, subdivisionsZ=1, name='cmbB')
			cmb = pm.polyUnite('cmbA', 'cmbB', ch=1, mergeUVSets=1, centerPivot=1, name='cmb')

		self.perform_test({
			"list(self.getIslands('cmb'))": [['cmb.f[0]', 'cmb.f[5]', 'cmb.f[4]', 'cmb.f[9]', 'cmb.f[6]', 'cmb.f[1]', 'cmb.f[10]', 'cmb.f[11]', 'cmb.f[14]', 'cmb.f[8]', 'cmb.f[7]', 'cmb.f[3]', 'cmb.f[13]', 'cmb.f[2]', 'cmb.f[12]'], ['cmb.f[15]', 'cmb.f[20]', 'cmb.f[19]', 'cmb.f[24]', 'cmb.f[21]', 'cmb.f[16]', 'cmb.f[25]', 'cmb.f[26]', 'cmb.f[29]', 'cmb.f[23]', 'cmb.f[22]', 'cmb.f[18]', 'cmb.f[28]', 'cmb.f[17]', 'cmb.f[27]']],
		})

	def test_getBorderComponents(self):
		'''
		'''
		self.perform_test({
			"self.getBorderComponents('pln', 'vtx')": ['plnShape.vtx[0:4]', 'plnShape.vtx[7:8]', 'plnShape.vtx[11:15]'],
			"self.getBorderComponents('pln', 'face')": ['plnShape.f[0:3]', 'plnShape.f[5:8]'],
			"self.getBorderComponents('pln')": ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
			"self.getBorderComponents('pln.e[:]')": ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
			"self.getBorderComponents(['pln.e[9]','pln.e[10]', 'pln.e[12]', 'pln.e[16]'], 'f', componentBorder=True)": ['plnShape.f[1]', 'plnShape.f[3:5]', 'plnShape.f[7]'],
			"self.getBorderComponents('pln.f[3:4]', 'f', componentBorder=True)": ['plnShape.f[0:1]', 'plnShape.f[5:7]'],
			"self.getBorderComponents('pln.f[3:4]', 'vtx', componentBorder=True)": ['plnShape.vtx[4:6]', 'plnShape.vtx[8:10]'],
			"self.getBorderComponents('pln.vtx[6]', 'e', componentBorder=True)": ['plnShape.e[5]', 'plnShape.e[9]', 'plnShape.e[11:12]'],
		})

	def test_getClosestVerts(self):
		'''
		'''
		self.perform_test({
			"self.getClosestVerts('pln.vtx[:10]', 'pln.vtx[11:]', 6.667)": [('plnShape.vtx[7]', 'plnShape.vtx[11]'), ('plnShape.vtx[8]', 'plnShape.vtx[12]'), ('plnShape.vtx[9]', 'plnShape.vtx[13]'), ('plnShape.vtx[10]', 'plnShape.vtx[11]'), ('plnShape.vtx[10]', 'plnShape.vtx[14]')],
		})

	def test_getClosestVertex(self):
		'''
		'''
		self.perform_test({
			"self.getClosestVertex('plnShape.vtx[0]', 'cyl', returnType='int')": {'plnShape.vtx[0]': 6},
			"self.getClosestVertex('plnShape.vtx[0]', 'cyl')": {'plnShape.vtx[0]': 'cylShape.vtx[6]'},
			"self.getClosestVertex('plnShape.vtx[2:3]', 'cyl')": {'plnShape.vtx[2]': 'cylShape.vtx[9]', 'plnShape.vtx[3]': 'cylShape.vtx[9]'},
		})

	def test_getEdgePath(self):
		'''
		'''
		self.perform_test({
			"self.getEdgePath('sph.e[12]', 'edgeLoop')": ['sphShape.e[12]', 'sphShape.e[17]', 'sphShape.e[16]', 'sphShape.e[15]', 'sphShape.e[14]', 'sphShape.e[13]'],
			"self.getEdgePath('sph.e[12]', 'edgeLoop', 'int')": [12, 17, 16, 15, 14, 13],
			"self.getEdgePath('sph.e[12]', 'edgeRing')": ['sphShape.e[0]', 'sphShape.e[6]', 'sphShape.e[12]', 'sphShape.e[18]', 'sphShape.e[24]'],
			"self.getEdgePath(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath')": ['sphShape.e[43]', 'sphShape.e[42]', 'sphShape.e[47]', 'sphShape.e[46]'],
			"self.getEdgePath(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath')": ['sphShape.e[60]', 'sphShape.e[48]', 'sphShape.e[42]', 'sphShape.e[36]', 'sphShape.e[30]', 'sphShape.e[54]'],
		})

	def test_getEdgesByNormalAngle(self):
		'''
		'''
		self.perform_test({
			"self.getEdgesByNormalAngle('cyl', 50, 130)": ['cylShape.e[0:23]'],
		})

	def test_getComponentsByNumberOfConnected(self):
		'''
		'''
		self.perform_test({
			"self.getComponentsByNumberOfConnected(['sph.f[18:23]', 'sph.f[30:35]'], 3, 'e')": ['sphShape.f[30]', 'sphShape.f[31]', 'sphShape.f[32]', 'sphShape.f[33]', 'sphShape.f[34]', 'sphShape.f[35]'],
			"self.getComponentsByNumberOfConnected('pln.vtx[:]', (0,2), 'e')": ['plnShape.vtx[0]', 'plnShape.vtx[3]', 'plnShape.vtx[12]', 'plnShape.vtx[15]'],
		})

	def test_getVertexNormal(self):
		'''
		'''
		import maya.api.OpenMaya as om
		self.perform_test({
			"self.getVertexNormal('pln.vtx[2]', angleWeighted=False)": om.MVector(0, 1, 0),
		})

	def test_getVectorFromComponents(self):
		'''
		'''
		self.perform_test({
			"self.getVectorFromComponents('pln.f[:]')": (0.0, 1.0, 0.0),
			"self.getVectorFromComponents(['cyl.f[7]', 'cyl.f[8]'])": (0.0, 0.0, 0.43982641180356435),
		})

# -----------------------------------------------------------------------------

if __name__=='__main__':

	unittest.main(exit=False)






# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})
# """

# # Deprecated ---------------------


# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from pythontk import Str, Iter
#from this package:
from mayatk.nodetk import Node


class Rig(object):
	'''
	'''
	@staticmethod
	def createLocator(name=None, pos=(), scale=1):
		'''Create a locator with the given scale.

		:Parameters:
			name (str) = Name the locator.
			pos (tuple) = The desired location in world space.
			scale (float) = The desired scale of the locator.

		:Return:
			(obj) locator
		'''
		loc = pm.spaceLocator()

		if name:
			pm.rename(loc, name)
		if pos:
			pm.move(pos[0], pos[1], pos[2], loc)
		if scale!=1:
			pm.scale(loc, scale, scale, scale) #scale the locator

		return loc


	@classmethod
	def removeLocator(cls, objects):
		'''Remove a parented locator from the child object.

		:Parameters:
			obj (str)(obj)(list) = The child object or the locator itself.
		'''
		errorMsg = lambda: pm.inViewMessage(
			statusMessage="{} in removeLocator\n\t# Error: Unable to remove locator for the given object. #".format(__file__), 
			pos='topCenter', 
			fade=True
		)

		pm.undoInfo(openChunk=1)

		for obj in pm.ls(objects, long=True, objectsOnly=True):
			if not pm.objExists(obj):
				continue

			elif Node.isLocator(obj) and not Node.getType(obj) and not Node.getChildren(obj):
				pm.delete(obj)
				continue

			#unlock attributes
			cls.setAttrLockState(obj, translate=False, rotate=False, scale=False) #unlock all.

			if not Node.isLocator(obj):
				try: #if the 'obj' is not a locator, check if it's parent is.
					obj = Node.getType(obj)
					if not Node.isLocator(obj):
						errorMsg()
						continue
				except IndexError as error:
					errorMsg()
					continue

			try: #remove from group
				grp = Node.getType(obj)
			except IndexError as error:
				errorMsg()
				continue
			if Node.isGroup(grp):
				if grp.split('_')[0]==obj.split('_')[0]:
					pm.ungroup(grp)

			#remove locator
			pm.ungroup(obj)
			pm.delete(obj)

		pm.undoInfo(closeChunk=1)


	@staticmethod
	def resetPivotTransforms(objects):
		'''Reset Pivot Transforms
		'''
		objs = pm.ls(type=('transform', 'geometryShape'), sl=1)

		if len(objs)>0:
			pm.xform(cp=1)
			
		pm.manipPivot(ro=1, rp=1)


	@staticmethod
	def bakeCustomPivot(objects, position=False, orientation=False):
		'''
		'''
		transforms = pm.ls(objects, transforms=1)
		shapes = pm.ls(objects, shapes=1)
		objects = transforms+pm.listRelatives(shapes, path=1, parent=1, type='transform')

		ctx = pm.currentCtx()
		pivotModeActive = 0
		customModeActive = 0
		if ctx in ('RotateSuperContext', 'manipRotateContext'): #Rotate tool
			customOri = pm.manipRotateContext('Rotate', q=1, orientAxes=1)
			pivotModeActive = pm.manipRotateContext('Rotate', q=1, editPivotMode=1)
			customModeActive = pm.manipRotateContext('Rotate', q=1, mode=1)==3
		elif ctx in ('scaleSuperContext', 'manipScaleContext'): #Scale tool
			customOri = pm.manipScaleContext('Scale', q=1, orientAxes=1)
			pivotModeActive = pm.manipScaleContext('Scale', q=1, editPivotMode=1)
			customModeActive = pm.manipScaleContext('Scale', q=1, mode=1)==6
		else: #use the move tool orientation
			customOri = pm.manipMoveContext('Move', q=1, orientAxes=1) #get custom orientation
			pivotModeActive = pm.manipMoveContext('Move', q=1, editPivotMode=1)
			customModeActive = pm.manipMoveContext('Move', q=1, mode=1)==6
			if not ctx in ('moveSuperContext', 'manipMoveContext'): #Move tool
				otherToolActive = 1 #some other tool 

		if orientation and customModeActive:
			if not position:
				pm.mel.error((pm.mel.uiRes("m_bakeCustomToolPivot.kWrongAxisOriToolError")))
				return

			from math import degrees

			cX, cY, cZ = customOri = [
				degrees(customOri[0]),
				degrees(customOri[1]),
				degrees(customOri[2]),
			]

			pm.rotate(objects, cX, cY, cZ, a=1, pcp=1, pgp=1, ws=1, fo=1) #Set object(s) rotation to the custom one (preserving child transform positions and geometry positions)

		if position:
			for obj in objects:
				#Get pivot in parent space
				m = pm.xform(obj, q=1, m=1)
				p = pm.xform(obj, q=1, os=1, sp=1)
				oldX, oldY, oldZ = old = [
					(p[0]*m[0] + p[1]*m[4]+ p[2]*m[8]  + m[12]),
					(p[0]*m[1] + p[1]*m[5]+ p[2]*m[9]  + m[13]),
					(p[0]*m[2] + p[1]*m[6]+ p[2]*m[10] + m[14]),
				]

				pm.xform(obj, zeroTransformPivots=1) #Zero out pivots

				#Translate obj(s) back to previous pivot (preserving child transform positions and geometry positions)
				newX, newY, newZ = new = pm.getAttr(obj.name() + '.translate') #obj.translate
				pm.move(obj, oldX-newX, oldY-newY, oldZ-newZ, pcp=1, pgp=1, ls=1, r=1)

		if pivotModeActive:
			pm.ctxEditMode() #exit pivot mode

		#Set the axis orientation mode back to obj
		if orientation and customModeActive:
			if ctx in ('RotateSuperContext', 'manipRotateContext'):
				pm.manipPivot(rotateToolOri=0)
			elif ctx in ('scaleSuperContext', 'manipScaleContext'):
				pm.manipPivot(scaleToolOri=0)
			else: #Some other tool #Set move tool to obj mode and clear the custom ori. (so the tool won't restore it when activated)
				pm.manipPivot(moveToolOri=0)
				if not ctx in ('moveSuperContext', 'manipMoveContext'):
					pm.manipPivot(ro=1)


	@classmethod
	def setAttrLockState(cls, objects, translate=None, rotate=None, scale=None, **kwargs):
		'''Lock/Unlock any attribute for the given objects, by passing it into kwargs as <attr>=<value>.
		A 'True' value locks the attribute, 'False' unlocks, while 'None' leaves the state unchanged.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to lock/unlock attributes of.
			translate (bool) = Lock/Unlock all translate x,y,z values at once.
			rotate (bool) = Lock/Unlock all rotate x,y,z values at once.
			scale (bool) = Lock/Unlock all scale x,y,z values at once.
		'''
		objects = pm.ls(objects, transforms=1, long=True)

		attrs_and_state = {
			('tx','ty','tz'):translate, #attributes and state. ex. ('tx','ty','tz'):False
			('rx','ry','rz'):rotate, 
			('sx','sy','sz'):scale
		}

		attrs_and_state.update(kwargs) #update the dict with any values from kwargs.

		for obj in objects:
			try:
				if Node.isLocator(obj):
					obj = pm.listRelatives(obj, children=1, type='transform')[0]
			except IndexError as error:
				return

			for attrs, state in attrs_and_state.items():
				if state is None:
					continue
				for a in Iter.makeList(attrs):
					pm.setAttr('{}.{}'.format(obj, a), lock=state)


	@staticmethod
	def createGroup(objects=[], name='', zeroTranslation=False, zeroRotation=False, zeroScale=False):
		'''Create a group containing any given objects.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to group.
			name (str) = Name the group.
			zeroTranslation (bool) = Freeze translation before parenting.
			zeroRotation (bool) = Freeze rotation before parenting.
			zeroScale (bool) = Freeze scale before parenting.

		:Return:
			(obj) the group.
		'''
		grp = pm.group(empty=True, n=name)
		try:
			pm.parent(grp, objects)
		except Exception as error:
			print ('{} in createGroup\n\t# Error: Unable to parent object(s): {} #'.format(__file__, error))

		if zeroTranslation:
			for attr in ('tx','ty','tz'):
				pm.setAttr(getattr(grp, attr), 0) #pm.setAttr(node.translate, 0)
		if zeroRotation:
			for attr in ('rx','ry','rz'):
				pm.setAttr(getattr(grp, attr), 0)
		if zeroScale:
			for attr in ('sx','sy','sz'):
				pm.setAttr(getattr(grp, attr), 0)

		pm.parent(grp, world=True)
		return grp


	@classmethod
	def createGroupLRA(cls, objects, name='', makeIdentity=True):
		'''Creates a group using the first object to define the local rotation axis.

		:Parameters:
			objects (str)(obj)(list) = The objects to group. The first object will be used to define the groups LRA.
			name (str) = The group name.
			makeIdentity (bool) = Freeze transforms on group child objects.
		'''
		try:
			obj, *other = pm.ls(objects, transforms=1)
		except IndexError as error:
			print('{} in createGroupLRA\n\t# Error: Operation requires at least one object. #'.format(__file__))
			return None

		pm.undoInfo(openChunk=1)
		cls.bakeCustomPivot(obj, position=True, orientation=True) #pm.mel.BakeCustomPivot(obj) #bake the pivot on the object that will define the LRA.

		grp = pm.group(empty=True)
		pm.parent(grp, obj)

		pm.setAttr(grp.translate, (0,0,0))
		pm.setAttr(grp.rotate, (0,0,0))

		objParent = pm.listRelatives(obj, parent=1)
		pm.parent(grp, objParent) #parent the instance under the original objects parent.

		try:
			pm.parent(obj, grp)
		except: #root level objects
			pm.parent(grp, world=True)
			pm.parent(obj, grp)

		for o in other: #parent any other objects to the new group.
			pm.parent(o, grp)
			if makeIdentity:
				pm.makeIdentity(o, apply=True) #freeze transforms on child objects.

		if not name and objParent: #name the group.
			pm.rename(grp, objParent[0].name())
		elif not name:
			pm.rename(grp, obj.name())
		else:
			pm.rename(grp, name)
		pm.undoInfo(closeChunk=1)

		return grp


	@classmethod
	def createLocatorAtObject(cls, objects, parent=False, freezeTransforms=False, bakeChildPivot=False, 
			grpSuffix='_GRP#', locSuffix='_LOC#', objSuffix='_GEO#', stripDigits=False, stripSuffix=False, 
			scale=1, lockTranslate=False, lockRotation=False, lockScale=False):
		'''Create locators with the same transforms as any selected object(s).
		If there are vertices selected it will create a locator at the center of the selected vertices bounding box.

		:Parameters:
			objects (str)(obj)(list) = A list of objects, or an object name to create locators at.
			parent (bool) = Parent the object to the locator. (default=False)
			freezeTransforms (bool) = Freeze transforms on the locator. (default=True)
			bakeChildPivot (bool) = Bake pivot positions on the child object. (default=True)
			grpSuffix (str) = A string appended to the end of the created groups name. (default: '_GRP#')
			locSuffix (str) = A string appended to the end of the created locators name. (default: '_LOC#')
			objSuffix (str) = A string appended to the end of the existing objects name. (default: '_GEO#')
			stripDigits (bool) = Strip numeric characters from the string. If the resulting name is not unique, maya will append a trailing digit. (default=False)
			stripSuffix (str) = Strip any existing suffix. A suffix is defined by the last '_' (if one exists) and any chars trailing. (default=False)
			scale (float) = The scale of the locator. (default=1)
			lockTranslate (bool) = Lock the translate values of the child object. (default=False)
			lockRotation (bool) = Lock the rotation values of the child object. (default=False)
			lockScale (bool) = Lock the scale values of the child object. (default=False)
			remove (bool) = Removes the locator and any child locks. (not valid with component selections) (default=False)

		ex. call: createLocatorAtSelection(strip='_GEO', suffix='', stripDigits=True, parent=True, lockTranslate=True, lockRotation=True)
		'''
		getSuffix = lambda o: locSuffix if Node.isLocator(o) else grpSuffix if Node.isGroup(o) else objSuffix #match the correct suffix to the object type.

		pm.undoInfo(openChunk=1)

		for obj in pm.ls(objects, long=True, type='transform'):

			if bakeChildPivot:
				cls.bakeCustomPivot(obj, position=1, orientation=1)

			vertices = pm.filterExpand(obj, sm=31) #returns a string list.
			if vertices:
				objName = vertices[0].split('.')[0]
				obj = pm.ls(objName)

				loc = cls.createLocator(scale=scale)

				xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(vertices)
				x, y, z = pos = ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
				pm.move(x, y, z, loc)

			else: #object:
				loc = cls.createLocator(scale=scale)
				tempConst = pm.parentConstraint(obj, loc, maintainOffset=False)
				pm.delete(tempConst)

			try:
				if parent:
					origParent = pm.listRelatives(obj, parent=1)

					grp = cls.createGroup(obj, zeroTranslation=1, zeroRotation=1)
					pm.rename(grp, Str.formatSuffix(obj.name(), suffix=getSuffix(grp), strip=(objSuffix, grpSuffix, locSuffix), stripTrailingInts=stripDigits, stripTrailingAlpha=stripSuffix))

					pm.parent(obj, loc)
					pm.parent(loc, grp)
					pm.parent(grp, origParent)

				if freezeTransforms: #freeze transforms before baking pivot.
					cls.setAttrLockState(obj, translate=False, rotate=False, scale=False) #assure attributes are unlocked.
					pm.makeIdentity(obj, apply=True, normal=1)
					pm.makeIdentity(loc, apply=True, normal=1) #1=the normals on polygonal objects will be frozen. 2=the normals on polygonal objects will be frozen only if its a non-rigid transformation matrix.

				pm.rename(loc, Str.formatSuffix(obj.name(), suffix=getSuffix(loc), strip=(objSuffix, grpSuffix, locSuffix), stripTrailingInts=stripDigits, stripTrailingAlpha=stripSuffix))
				pm.rename(obj, Str.formatSuffix(obj.name(), suffix=getSuffix(obj), strip=(objSuffix, grpSuffix, locSuffix), stripTrailingInts=stripDigits, stripTrailingAlpha=stripSuffix))

				cls.setAttrLockState(obj, translate=lockTranslate, rotate=lockRotation, scale=lockScale)

			except Exception as error:
				pm.delete(loc)
				raise (error)

		pm.undoInfo(closeChunk=1)

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Return:
		(obj)
	'''
	try:
		return getattr(Rig, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')




# --------------------------------------------------------------------------------------------

if __name__=='__main__':

	sel = pm.ls(sl=1)

	Rigtk.createLocatorAtObject(sel,
		parent=1,
		freezeTransforms=1,
		bakeChildPivot=1,
		grpSuffix='_GRP',
		locSuffix='_LCTR',
		objSuffix='_GEO',
		stripDigits=1,
		stripSuffix=1,
		scale=1,
		lockTranslate=0,
		lockRotation=0,
		lockScale=0,
	)

# --------------------------------------------------------------------------------------------









# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------



# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (f'# Error: {__file__}: {error}')

from pythontk import Iter


def undo(fn):
	'''A decorator to place a function into Maya's undo chunk.
	Prevents the undo queue from breaking entirely if an exception is raised within the given function.

	:Parameters:
		fn (obj) = The decorated python function that will be placed into the undo que as a single entry.
	'''
	def wrapper(*args, **kwargs):
		with pm.UndoChunk():
			rtn = fn(*args, **kwargs)
			return rtn
	return wrapper


def getMainWindow():
	'''Get maya's main window object.

	:Return:
		(QWidget)
	'''
	from PySide2.QtWidgets import QApplication

	app = QApplication.instance()
	if not app:
		print ('{} in getMainWindow\n\t# Warning: Could not find QApplication instance. #'.format(__file__))
		return None

	main_window = next(iter(w for w in app.topLevelWidgets() if w.objectName()=='MayaWindow'), None)
	if not main_window:
		print ('{} in getMainWindow\n\t# Warning: Could not find main window instance. #'.format(__file__))
		return None

	return main_window


def wrapControl(controlName, container):
	'''Embed a Maya Native UI Object.

	:Parameters:
		controlName (str) = The name of an existing maya control. ie. 'cmdScrollFieldReporter1'
		container (obj) = A widget instance in which to wrap the control.

	ex. call:
	modelPanelName = pm.modelPanel("embeddedModelPanel#", cam='persp')
	wrapControl(modelPanelName, QtWidgets.QtWidget())
	'''
	from PySide2 import QtWidgets
	from shiboken2 import wrapInstance
	from maya.OpenMayaUI import MQtUtil

	layout = QtWidgets.QVBoxLayout(container)
	layout.setContentsMargins(0,0,0,0)
	layoutName = Str.setCase(container.objectName()+'Layout', 'camelCase') # results in '<objectName>Layout' or 'layout' if container objectName is ''
	layout.setObjectName(layoutName)
	pm.setParent(layoutName)

	from tentacle.switchboard import Switchboard
	derivedClass = Switchboard.getDerivedType(container)

	ptr = MQtUtil.findControl(controlName) #get a pointer to the maya api paneLayout.
	control = wrapInstance(int(ptr), derivedClass)
	layout.addWidget(control)

	return control


def mfnMeshGenerator(objects):
	'''Generate mfn mesh from the given list of objects.

	:Parameters:
		objects (str)(obj(list) = The objects to convert to mfn mesh.

	:Return:
		(generator)
	'''
	import maya.OpenMaya as om

	selectionList = om.MSelectionList()
	for mesh in getShapeNode(pm.ls(objects)):
		selectionList.add(mesh)

	for i in range(selectionList.length()):    
		dagPath = om.MDagPath()
		selectionList.getDagPath(i, dagPath)
		# print (dagPath.fullPathName()) #debug
		mfnMesh = om.MFnMesh(dagPath)
		yield mfnMesh


def getArrayType(lst):
	'''Determine if the given element(s) type.
	Samples only the first element.

	:Parameters:
		obj (str)(obj)(list) = The components(s) to query.

	:Return:
		(list) 'str', 'obj'(shape node), 'transform'(as string), 'int'(valid only at sub-object level)

	ex. call:
	getArrayType('cyl.vtx[0]') #returns: 'transform'
	getArrayType('cylShape.vtx[:]') #returns: 'str'
	'''
	try:
		o = Iter.makeList(lst)[0]
	except IndexError as error:
		# print ('{}\n# Error: getArrayType: Operation requires at least one object. #\n	{}'.format(__file__, error))
		return ''

	return 'str' if isinstance(o, str) else 'int' if isinstance(o, int) else 'obj'


def convertArrayType(lst, returnType='str', flatten=False):
	'''Convert the given element(s) to <obj>, 'str', or int values.

	:Parameters:
		lst (str)(obj)(list) = The components(s) to convert.
		returnType (str) = The desired returned array element type.
			valid: 'str'(default), 'obj', 'int'(valid only at sub-object level).
		flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

	:Return:
		(list)(dict) return a dict only with a return type of 'int' and more that one object given.

	ex. call:
	convertArrayType('obj.vtx[:2]', 'str') #returns: ['objShape.vtx[0:2]']
	convertArrayType('obj.vtx[:2]', 'str', True) #returns: ['objShape.vtx[0]', 'objShape.vtx[1]', 'objShape.vtx[2]']
	convertArrayType('obj.vtx[:2]', 'obj') #returns: [MeshVertex('objShape.vtx[0:2]')]
	convertArrayType('obj.vtx[:2]', 'obj', True) #returns: [MeshVertex('objShape.vtx[0]'), MeshVertex('objShape.vtx[1]'), MeshVertex('objShape.vtx[2]')]
	convertArrayType('obj.vtx[:2]', 'int')) #returns: {nt.Mesh('objShape'): [(0, 2)]}
	convertArrayType('obj.vtx[:2]', 'int', True)) #returns: {nt.Mesh('objShape'): [0, 1, 2]}
	'''
	lst = pm.ls(lst, flatten=flatten)
	if not lst or isinstance(lst[0], int):
		return []

	if returnType=='int':
		result={}
		for c in lst:
			obj = pm.ls(c, objectsOnly=1)[0]
			num = c.split('[')[-1].rstrip(']')

			try:
				if flatten:
					componentNum = int(num)
				else:
					n = [int(n) for n in num.split(':')]
					componentNum = tuple(n) if len(n)>1 else n[0]

				if obj in result: #append to existing object key.
					result[obj].append(componentNum)
				else:
					result[obj] = [componentNum]
			except ValueError as error: #incompatible object type.
				print ('{} in convertArrayType\n\t# Error: unable to convert {} {} to int. #\n\t{}'.format(__file__, obj, num, error))
				break

		objects = set(pm.ls(lst, objectsOnly=True))
		if len(objects)==1: #flatten the dict values from 'result' and remove any duplicates.
			flattened = Iter.flatten(result.values())
			result = Iter.removeDuplicates(flattened)

	elif returnType=='str':
		result = list(map(str, lst))

	else:
		result = lst

	return result


def getParameterValuesMEL(node, cmd, parameters):
	'''Query a Maya command, and return a key:value pair for each of the given parameters.

	:Parameters:
		node (str)(obj)(list) = The object to query attributes of.
		parameters (list) = The command parameters to query. ie. ['enableTranslationX','translationX']

	:Return:
		(dict) {'parameter name':<value>} ie. {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]}

	ex. call: getParameterValuesMEL(obj, 'transformLimits', ['enableTranslationX','translationX'])
	'''
	cmd = getattr(pm, cmd)
	node = pm.ls(node)[0]

	result={}
	for p in parameters:
		values = cmd(node, **{'q':True, p:True}) #query the parameter to get it's value.

		result[p] = values

	return result


def setParameterValuesMEL(node, cmd, parameters):
	'''Set parameters using a maya command.

	:Parameters:
		node (str)(obj)(list) = The object to query attributes of.
		parameters (dict) = The command's parameters and their desired values. ie. {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]}

	ex. call: setParameterValuesMEL(obj, 'transformLimits', {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]})
	'''
	cmd = getattr(pm, cmd)
	node = pm.ls(node)[0]

	for p, v in parameters.items():
		cmd(node, **{p:v})


def getSelectedChannels():
	'''Get any attributes (channels) that are selected in the channel box.

	:Return:
		(str) list of any selected attributes as strings. (ie. ['tx', ry', 'sz'])
	'''
	channelBox = pm.mel.eval('global string $gChannelBoxName; $temp=$gChannelBoxName;') #fetch maya's main channelbox
	attrs = pm.channelBox(channelBox, q=True, sma=True)

	if attrs is None:
		attrs=[]
	return attrs


def getPanel(*args, **kwargs):
	'''Returns panel and panel configuration information.
	A fix for the broken pymel command of the same name.

	:Parameters:
		[allConfigs=boolean], [allPanels=boolean], [allScriptedTypes=boolean], [allTypes=boolean], [configWithLabel=string], [containing=string], [invisiblePanels=boolean], [scriptType=string], [type=string], [typeOf=string], [underPointer=boolean], [visiblePanels=boolean], [withFocus=boolean], [withLabel=string])

	:Return:
		(str) An array of panel names.
	'''
	from maya.cmds import getPanel #pymel getPanel is broken in ver: 2022.

	result = getPanel(*args, **kwargs)

	return result


def mainProgressBar(size, name="progressBar#", stepAmount=1):
	'''#add esc key pressed return False

	:Parameters:
		size (int) = total amount
		name (str) = name of progress bar created
		stepAmount(int) = increment amount

	example use-case:
	mainProgressBar (len(edges), progressCount)
		pm.progressBar ("progressBar_", edit=1, step=1)
		if pm.progressBar ("progressBar_", query=1, isCancelled=1):
			break
	pm.progressBar ("progressBar_", edit=1, endProgress=1)

	to use main progressBar: name=string $gMainProgressBar
	'''
	status = 'processing: {} items ..'.format(size)

	edit=False
	if pm.progressBar(name, exists=1):
		edit=True

	pm.progressBar(name, edit=edit,
					beginProgress=1,
					isInterruptable=True,
					status=status,
					maxValue=size,
					step=stepAmount)


def viewportMessage(message='', statusMessage='', assistMessage='', position='topCenter'):
	'''
	:Parameters:
		message (str) = The message to be displayed, (accepts html formatting). General message, inherited by -amg/assistMessage and -smg/statusMessage.
		statusMessage (str) = The status info message to be displayed (accepts html formatting).
		assistMessage (str) = The user assistance message to be displayed, (accepts html formatting).
		position (str) = position on screen. possible values are: topCenter","topRight","midLeft","midCenter","midCenterTop","midCenterBot","midRight","botLeft","botCenter","botRight"

	ex. viewportMessage("shutting down:<hl>"+str(timer)+"</hl>")
	'''
	fontSize=10
	fade=1
	fadeInTime=0
	fadeStayTime=1000
	fadeOutTime=500
	alpha=75

	if message:
		pm.inViewMessage(message=message, position=position, fontSize=fontSize, fade=fade, fadeInTime=fadeInTime, fadeStayTime=fadeStayTime, fadeOutTime=fadeOutTime, alpha=alpha) #1000ms = 1 sec
	elif statusMessage:
		pm.inViewMessage(statusMessage=statusMessage, position=position, fontSize=fontSize, fade=fade, fadeInTime=fadeInTime, fadeStayTime=fadeStayTime, fadeOutTime=fadeOutTime, alpha=alpha) #1000ms = 1 sec
	elif assistMessage:
		pm.inViewMessage(assistMessage=assistMessage, position=position, fontSize=fontSize, fade=fade, fadeInTime=fadeInTime, fadeStayTime=fadeStayTime, fadeOutTime=fadeOutTime, alpha=alpha) #1000ms = 1 sec


def outputText (text, window_title):
	'''output text
	'''
	#window_title = pm.mel.eval(python("window_title"))
	window = str(pm.window(	widthHeight=(300, 300), 
							topLeftCorner=(65,265),
							maximizeButton=False,
							resizeToFitChildren=True,
							toolbox=True,
							title=window_title))
	scrollLayout = str(pm.scrollLayout(verticalScrollBarThickness=16, 
									horizontalScrollBarThickness=16))
	pm.columnLayout(adjustableColumn=True)
	text_field = str(pm.text(label=text, align='left'))
	print(text_field)
	pm.setParent('..')
	pm.showWindow(window)
	return

# #output textfield parsed by ';'
# def outputTextField2(text):
# 	window = str(pm.window(	widthHeight=(250, 650), 
# 							topLeftCorner=(50,275),
# 							maximizeButton=False,
# 							resizeToFitChildren=False,
# 							toolbox=True,
# 							title=""))
# 	scrollLayout = str(pm.scrollLayout(verticalScrollBarThickness=16, 
# 									horizontalScrollBarThickness=16))
# 	pm.columnLayout(adjustableColumn=True)
# 	print(text)
# 	#for item in array:
# 	text_field = str(pm.textField(height=20,
# 										width=250, 
# 										editable=False,
# 										insertText=str(text)))
# 	pm.setParent('..')
# 	pm.showWindow(window)
# 	return


def outputscrollField (text, window_title, width, height):
	'''Create an output scroll layout.
	'''
	window_width  = width  * 300
	window_height = height * 600
	scroll_width  = width  * 294
	scroll_height = height * 590
	window = str(pm.window(	widthHeight=(window_width, window_height),
							topLeftCorner=(45, 0),
							maximizeButton=False,
							sizeable=False,
							title=window_title
							))
	scrollLayout = str(pm.scrollLayout(verticalScrollBarThickness=16, 
									horizontalScrollBarThickness=16))
	pm.columnLayout(adjustableColumn=True)
	scroll_field = str(pm.scrollField(text=(text),
									width=scroll_width,
									height=scroll_height,))
	print(window)
	pm.setParent('..')
	pm.showWindow(window)
	return scroll_field


def outputTextField (array, window_title):
	'''Create an output text field.
	'''
	window = str(pm.window(	widthHeight=(250, 650), 
							topLeftCorner=(65,275),
							maximizeButton=False,
							resizeToFitChildren=False,
							toolbox=True,
							title=window_title))
	scrollLayout = str(pm.scrollLayout(verticalScrollBarThickness=16, 
									horizontalScrollBarThickness=16))
	pm.columnLayout(adjustableColumn=True)
	for item in array:
		text_field = str(pm.textField(height=20,
										width=500, 
										editable=False,
										insertText=str(item)))
	pm.setParent('..')
	pm.showWindow(window)
	return text_field

# --------------------------------------------------------------------------------------------






# print (__package__, __file__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect

import pymel.core as pm

from mayatk import *
import test


sfr = pm.melGlobals['cmdScrollFieldReporter']
pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)


class Xformtk_test(Main, Xformtk):
	'''
	'''
	# Tear down the any previous test by creating a new scene:
	pm.mel.file(new=True, force=True)

	# assemble the test scene:
	if not pm.objExists('cube1'):
		cube1 = pm.polyCube(width=5, height=5, depth=5, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1, name='cube1')

	if not pm.objExists('cube2'):
		cube2 = pm.polyCube(width=2, height=4, depth=8, subdivisionsX=3, subdivisionsY=3, subdivisionsZ=3, name='cube2')

	if not pm.objExists('sph'):
		sph = pm.polySphere(radius=5, subdivisionsX=12, subdivisionsY=12, name='sph')

	def test_moveTo(self):
		'''
		'''
		self.perform_test({
			"self.moveTo('cube1', 'cube2')": None,
		})

	def test_dropToGrid(self):
		'''
		'''
		self.perform_test({
			"self.dropToGrid('cube1', align='Min', origin=True, centerPivot=True, freezeTransforms=True)": None,
		})

	def test_resetTranslation(self):
		'''
		'''
		self.perform_test({
			"self.resetTranslation('cube1')": None,
		})

	def test_setTranslationToPivot(self):
		'''
		'''
		self.perform_test({
			"self.setTranslationToPivot('cube1')": None,
		})

	def test_alignPivotToSelection(self):
		'''
		'''
		self.perform_test({
			"self.alignPivotToSelection('cube1', 'cube2')": None,
		})

	def test_aimObjectAtPoint(self):
		'''
		'''
		self.perform_test({
			"self.aimObjectAtPoint(['cube1', 'cube2'], (0, 15, 15))": None,
		})

	def test_rotateAxis(self):
		'''
		'''
		self.perform_test({
			"self.rotateAxis(['cube1', 'cube2'], (0, 15, 15))": None,
		})

	def test_getOrientation(self):
		'''
		'''
		self.perform_test({
			"self.getOrientation('cube1')": ([1, 0, 0], [0, 1, 0], [0, 0, 1]),
		})

	def test_getDistanceBetweenTwoObjects(self):
		'''
		'''
		self.dropToGrid(['cube1', 'cube2'], origin=True, centerPivot=True)
		pm.move('cube2', 0, 0, 15)

		self.perform_test({
			"self.getDistanceBetweenTwoObjects('cube1', 'cube2')": 15,
		})

	def test_getCenterPoint(self):
		'''
		'''
		self.perform_test({
			"self.getCenterPoint('sph')": (0, 0, 0),
			"self.getCenterPoint('sph.vtx[*]')": (0, 0, 0),
		})


	def test_getBoundingBoxValue(self):
		'''
		'''
		self.perform_test({
			"self.getBoundingBoxValue('sph')": (10, 10, 10),
		})

	def test_sortByBoundingBoxValue(self):
		'''
		'''
		self.perform_test({
			"str(self.sortByBoundingBoxValue(['sph.vtx[0]', 'sph.f[0]']))": "[MeshFace('sphShape.f[0]'), MeshVertex('sphShape.vtx[0]')]",
		})

	def test_matchScale(self):
		'''
		'''
		self.perform_test({
			"self.matchScale('cube1', 'cube2', scale=False)": [1.3063946090989371, 0.539387725343009, 0.539387708993454],
		})

	def test_snap3PointsTo3Points(self):
		'''
		'''
		self.perform_test({
			# "self.snap3PointsTo3Points()": None,
		})

	def test_isOverlapping(self):
		'''
		'''
		self.perform_test({
			# "self.isOverlapping()": None,
		})

	def test_alignVertices(self):
		'''
		'''
		self.perform_test({
			# "self.alignVertices()": None,
		})

# -----------------------------------------------------------------------------

if __name__=='__main__':

	unittest.main(exit=False)






# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})
# """

# # Deprecated ---------------------


# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from pythontk import Str, Iter, areSimilar
#from this package:
from mayatk.core import mfnMeshGenerator, viewportMessage, undo
from mayatk.nodetk import Node
from mayatk.cmpttk import Cmpt
from mayatk.xformtk import Xform


class Edit():
	'''
	'''
	@staticmethod
	@undo
	def rename(objects, to, fltr='', regEx=False, ignoreCase=False):
		'''Rename scene objects.

		:Parameters:
			objects (str)(obj)(list) = The object(s to rename. If nothing is given, all scene objects will be renamed.
			to (str) = Desired name: An optional asterisk modifier can be used for formatting
				chars - replace all.
				*chars* - replace only.
				*chars - replace suffix.
				**chars - append suffix.
				chars* - replace prefix.
				chars** - append prefix.
			fltr (str) = Optionally, filter which the given objects to rename using the following: 
				An asterisk denotes startswith*, *endswith, *contains*, and multiple search strings can be separated by pipe ('|') chars.
				chars - Search exact.
				*chars* - Search contains chars.
				*chars - Search endswith chars.
				chars* - Search startswith chars.
				chars|chars - Search any of.  can be used in conjuction with other modifiers.
			regEx (bool) = If True, regular expression syntax is used instead of the default '*' and '|' modifiers.
			ignoreCase (bool) = Ignore case when searching. Applies only to the 'fltr' parameter's search.

		ex. rename(r'Cube', '*001', regEx=True) #replace chars after 'fltr' on any object with a name that contains 'Cube'. ie. 'polyCube001' from 'polyCube'
		ex. rename(r'Cube', '**001', regEx=True) #append chars on any object with a name that contains 'Cube'. ie. 'polyCube1001' from 'polyCube1'
		'''
		# pm.undoInfo (openChunk=1)
		objects = pm.ls(objectsOnly=1) if not objects else pm.ls(objects)

		#get the short names from the long in order to correctly format. ex. 'NUT_' from: 'CENTER_HINGE_FEMALE_GRP|NUT_'
		long_names = [obj.name() for obj in objects]
		short_names = [ii if ii else i for i, ii in strtk.splitAtChars(long_names)] #split the long names at the last '|' to get the short name.

		names = strtk.findStrAndFormat(short_names, to, fltr, regEx=regEx, ignoreCase=ignoreCase, returnOldNames=True)
		print ('# Rename: Found {} matches. #'.format(len(names)))

		for i, (oldName, newName) in enumerate(names):
			oldName = long_names[i] #use the long name to reference the object instead.
			try:
				if pm.objExists(oldName):
					n = pm.rename(oldName, newName) #Rename the object with the new name
					if not n==newName:
						print ('# Warning: Attempt to rename "{}" to "{}" failed. Renamed instead to "{}". #'.format(oldName, newName, n))
					else:
						print ('# Result: Successfully renamed "{}" to "{}". #'.format(oldName, newName))

			except Exception as e:
				if not pm.ls(oldName, readOnly=True)==[]: #ignore read-only errors.
					print ('# Error: Attempt to rename "{}" to "{}" failed. {} #'.format(oldName, newName, str(e).rstrip()))
		# pm.undoInfo (closeChunk=1)


	@staticmethod
	@undo
	def setCase(objects=[], case='caplitalize'):
		'''Rename objects following the given case.

		:Parameters:
			objects (str)(list) = The objects to rename. default:all scene objects
			case (str) = Desired case using python case operators. 
				valid: 'upper', 'lower', 'caplitalize', 'swapcase' 'title'. default:'caplitalize'

		ex. call: setCase(pm.ls(sl=1), 'upper')
		'''
		# pm.undoInfo(openChunk=1)
		for obj in pm.ls(objects):
			name = obj.name()

			newName = getattr(name, case)()
			try:
				pm.rename(name, newName)
			except Exception as error:
				if not pm.ls(obj, readOnly=True)==[]: #ignore read-only errors.
					print (name+': ', error)
		# pm.undoInfo(closeChunk=1)


	@staticmethod
	@undo
	def setSuffixByObjLocation(objects, alphanumeric=False, stripTrailingInts=True, stripTrailingAlpha=True, reverse=False):
		'''Rename objects with a suffix defined by its location from origin.

		:Parameters:
			objects (str)(int)(list) = The object(s) to rename.
			alphanumeric (str) = When True use an alphanumeric character as a suffix when there is less than 26 objects else use integers.
			stripTrailingInts (bool) = Strip any trailing integers. ie. 'cube123'
			stripTrailingAlpha (bool) = Strip any trailing uppercase alphanumeric chars that are prefixed with an underscore.  ie. 'cube_A'
			reverse (bool) = Reverse the naming order. (Farthest object first)
		'''
		import string
		import re

		length = len(objects)
		if alphanumeric:
			if length<=26:
				suffix = string.ascii_lowercase.upper()
		else:
			suffix = [str(n).zfill(len(str(length))) for n in range(length)]

		ordered_objs = Xform.orderByDistance(objects, reverse=reverse)

		newNames={} #the object with the new name set as a key.
		for n, obj in enumerate(ordered_objs):

			current_name = obj.name()

			while ((current_name[-1]=='_' or current_name[-1].isdigit()) and stripTrailingInts) or ((len(current_name)>1 and current_name[-2]=='_' and current_name[-1].isupper()) and stripTrailingAlpha):
				if (current_name[-2]=='_' and current_name[-1].isupper()) and stripTrailingAlpha: #trailing underscore and uppercase alphanumeric char.
					current_name = re.sub(re.escape(current_name[-2:]) + '$', '', current_name)

				if (current_name[-1]=='_' or current_name[-1].isdigit()) and stripTrailingInts: #trailing underscore and integers.
					current_name = re.sub(re.escape(current_name[-1:]) + '$', '', current_name)

			newNames[obj] = current_name+'_'+suffix[n]

		for obj in ordered_objs: #rename all with a placeholder first so that there are no conflicts.
			pm.rename(obj, 'p0000000000')
		for obj in ordered_objs: #rename all with the new names.
			pm.rename(obj, newNames[obj])


	@staticmethod
	def snapClosestVerts(obj1, obj2, tolerance=10.0, freezeTransforms=False):
		'''Snap the vertices from object one to the closest verts on object two.

		:Parameters:
			obj1 (obj) = The object in which the vertices are moved from.
			obj2 (obj) = The object in which the vertices are moved to.
			tolerance (float) = Maximum search distance.
			freezeTransforms (bool) = Reset the selected transform and all of its children down to the shape level.
		'''
		vertices = Cmpt.getComponents(obj1, 'vertices')
		closestVerts = Cmpt.getClosestVertex(vertices, obj2, tolerance=tolerance, freezeTransforms=freezeTransforms)

		progressBar = "$gMainProgressBar"
		pm.progressBar(progressBar, edit=True, beginProgress=True, isInterruptable=True, status="Snapping Vertices ...", maxValue=len(closestVerts)) 

		pm.undoInfo(openChunk=True)
		for v1, v2 in closestVerts.items():
			if pm.progressBar(progressBar, query=True, isCancelled=True):
				break

			v2Pos = pm.pointPosition(v2, world=True)
			pm.xform(v1, translation=v2Pos, worldSpace=True)

			pm.progressBar(progressBar, edit=True, step=1)
		pm.undoInfo(closeChunk=True)

		pm.progressBar(progressBar, edit=True, endProgress=True)


	@staticmethod
	def mergeVertices(objects, selected=False, tolerance=0.001):
		'''Merge Vertices on the given objects.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to merge vertices on.
			selected (bool) = Merge only the currently selected components.
			tolerance (float) = The maximum merge distance.
		'''
		for obj in pm.ls(objects):

			if selected: #merge selected components.
				if pm.filterExpand(selectionMask=31): #selectionMask=vertices
					sel = pm.ls(obj, sl=1)
					pm.polyMergeVertex(sel, distance=tolerance, alwaysMergeTwoVertices=True, constructionHistory=True)
				else: #if selection type =edges or facets:
					pm.mel.MergeToCenter()

			else: #merge all vertices on the object.
				vertices = obj.vtx[:] # mel expression: select -r geometry.vtx[0:1135];
				pm.polyMergeVertex(vertices, distance=tolerance, alwaysMergeTwoVertices=False, constructionHistory=False)
				#return to original state
				pm.select(clear=1)
				pm.select(objects)


	@staticmethod
	def getAllFacesOnAxis(obj, axis="-x", localspace=False):
		'''Get all faces on a specified axis.

		:Parameters:
			obj (str)(obj) = The name of the geometry.
			axis (str) = The representing axis. case insensitive. (valid: 'x', '-x', 'y', '-y', 'z', '-z')
			localspace (bool) = Specify world or local space.

		ex call: getAllFacesOnAxis('polyObject', 'y')
		'''
		axis = axis.lower() #assure case.

		i=0 #'x'
		if any ([axis=="y",axis=="-y"]):
			i=1
		if any ([axis=="z",axis=="-z"]):
			i=2

		objName = pm.ls(obj)[0].name()

		if axis.startswith('-'): #any([axis=="-x", axis=="-y", axis=="-z"]):
			return list(face for face in pm.filterExpand(objName+'.f[*]', sm=34) if pm.exactWorldBoundingBox(face)[i] < -0.00001)
		else:
			return list(face for face in pm.filterExpand(objName+'.f[*]', sm=34) if pm.exactWorldBoundingBox(face)[i] > -0.00001)


	@classmethod
	def deleteAlongAxis(cls, obj, axis="-x"):
		'''Delete components of the given mesh object along the specified axis.

		:Parameters:
			obj (obj) = Mesh object.
			axis (str) = Axis to delete on. ie. '-x' Components belonging to the mesh object given in the 'obj' arg, that fall on this axis, will be deleted. 
		'''
		from mayatk import viewportMessage

		for node in [n for n in pm.listRelatives(obj, allDescendents=1) if pm.objectType(n, isType='mesh')]: #get any mesh type child nodes of obj.
			faces = cls.getAllFacesOnAxis(node, axis)
			if len(faces)==pm.polyEvaluate(node, face=1): #if all faces fall on the specified axis.
				pm.delete(node) #delete entire node
			else:
				pm.delete(faces) #else, delete any individual faces.

		viewportMessage("Delete faces on <hl>"+axis.upper()+"</hl>.")


	@classmethod
	def cleanGeometry(cls, objects, allMeshes=False, repair=False, quads=False, nsided=False, concave=False, holed=False, nonplanar=False, 
					zeroGeom=False, zeroGeomTol=0.000010, zeroEdge=False, zeroEdgeTol=0.000010, zeroMap=False, zeroMapTol=0.000010, 
					sharedUVs=False, nonmanifold=False, lamina=False, invalidComponents=False, splitNonManifoldVertex=False, historyOn=True):
		'''Select or remove unwanted geometry from a polygon mesh.

		:Parameters:
			objects (str)(obj)(list) = The polygon objects to clean.
			allMeshes (bool) = Clean all geomtry in the scene instead of only the current selection.
			repair (bool) = Attempt to repair instead of just selecting geometry.
		'''
		arg_list = '"{0}","{1}","{2}","{3}","{4}","{5}","{6}","{7}","{8}","{9}","{10}","{11}","{12}","{13}","{14}","{15}","{16}","{17}"'.format(
				allMeshes, 1 if repair else 2, historyOn, quads, nsided, concave, holed, nonplanar, zeroGeom, zeroGeomTol, 
				zeroEdge, zeroEdgeTol, zeroMap, zeroMapTol, sharedUVs, nonmanifold, lamina, invalidComponents)
		command = 'polyCleanupArgList 4 {'+arg_list+'}' # command = 'polyCleanup '+arg_list #(not used because of arg count error, also the quotes in the arg list would need to be removed). 

		if splitNonManifoldVertex: #Split Non-Manifold Vertex
			nonManifoldVerts = cls.findNonManifoldVertex(objects, select=2) #Select: 0=off, 1=on, 2=on while keeping any existing vertex selections. (default: 1)
			if repair:
				for vertex in nonManifoldVerts:
					cls.splitNonManifoldVertex(vertex, select=True) #select(bool): Select the vertex after the operation. (default: True)

		pm.select(objects)
		pm.mel.eval(command); #print (command)


	@staticmethod
	def getOverlappingDupObjects(objects=[], omitInitialObjects=False, select=False, verbose=False):
		'''Find any duplicate overlapping geometry at the object level.

		:Parameters:
			objects (list) = A list of objects to find duplicate overlapping geometry for. Default is selected objects, or all if nothing is selected.
			omitInitialObjects (bool) = Search only for duplicates of the given objects (or any selected objects if None given), and omit them from the return results.
			select (bool) = Select any found duplicate objects.
			verbose (bool) = Print each found object to console.

		:Return:
			(set)

		ex call: duplicates = getOverlappingDupObjects(omitInitialObjects=True, select=True, verbose=True)
		'''
		scene_objs = pm.ls(transforms=1, geometry=1) #get all scene geometry

		#attach a unique identifier consisting each objects polyEvaluate attributes, and it's bounding box center point in world space.
		scene_objs = {i:str(pm.objectCenter(i))+str(pm.polyEvaluate(i)) for i in scene_objs if not Node.isGroup(i)}
		selected_objs = pm.ls(scene_objs.keys(), sl=1) if not objects else objects

		objs_inverted={} #invert the dict, combining objects with like identifiers.
		for k, v in scene_objs.items():
			objs_inverted[v] = objs_inverted.get(v, []) + [k]

		duplicates=set()
		for k, v in objs_inverted.items():
			if len(v)>1:
				if selected_objs: #limit scope to only selected objects.
					if set(selected_objs) & set(v): #if any selected objects in found duplicates:
						if omitInitialObjects:
							[duplicates.add(i) for i in v if i not in selected_objs] #add any duplicated of that object, omitting the selected object.
						else:
							[duplicates.add(i) for i in v[1:]] #add all but the first object to the set of duplicates.
				else:
					[duplicates.add(i) for i in v[1:]] #add all but the first object to the set of duplicates.

		if verbose:
			for i in duplicates:
				print ('# Found: overlapping duplicate object: {} #'.format(i))
		print ('# {} overlapping duplicate objects found. #'.format(len(duplicates)))

		if select:
			pm.select(duplicates)

		return duplicates


	@staticmethod
	def findNonManifoldVertex(objects, select=1):
		'''Locate a connected vertex of non-manifold geometry where the faces share a single vertex.

		:Parameters:
			objects (str)(obj)(list) = A polygon mesh, or a list of meshes.
			select (int) = Select any found non-manifold vertices. 0=off, 1=on, 2=on while keeping any existing vertex selections. (default: 1)

		:Return:
			(set) any found non-manifold verts.
		'''
		pm.undoInfo(openChunk=True)
		nonManifoldVerts=set()

		vertices = Cmpt.getComponents(objects, 'vertices')
		for vertex in vertices:

			connected_faces = pm.polyListComponentConversion(vertex, fromVertex=1, toFace=1) #pm.mel.PolySelectConvert(1) #convert to faces
			connected_faces_flat = pm.ls(connected_faces, flatten=1) #selectedFaces = pm.ls(sl=1, flatten=1)

			#get a list of the edges of each face that is connected to the original vertex.
			edges_sorted_by_face=[]
			for face in connected_faces_flat:

				connected_edges = pm.polyListComponentConversion(face, fromFace=1, toEdge=1) #pm.mel.PolySelectConvert(1) #convert to faces
				connected_edges_flat = [str(i) for i in pm.ls(connected_edges, flatten=1)] #selectedFaces = pm.ls(sl=1, flatten=1)
				edges_sorted_by_face.append(connected_edges_flat)

			out=[] #1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
			while len(edges_sorted_by_face)>0:
				first, rest = edges_sorted_by_face[0], edges_sorted_by_face[1:] #first list, all other lists, of the list of lists.
				first = set(first)

				lf = -1
				while len(first)>lf:
					lf = len(first)

					rest2=[]
					for r in rest:
						if len(first.intersection(set(r)))>0:
							first |= set(r)
						else:
							rest2.append(r)     
					rest = rest2

				out.append(first)
				edges_sorted_by_face = rest

			if len(out)>1:
				nonManifoldVerts.add(vertex)
		pm.undoInfo(closeChunk=True)

		if select==2:
			pm.select(nonManifoldVerts, add=1)
		elif select==1:
			pm.select(nonManifoldVerts)

		return nonManifoldVerts


	@staticmethod
	def splitNonManifoldVertex(vertex, select=True):
		'''Separate a connected vertex of non-manifold geometry where the faces share a single vertex.

		:Parameters:
			vertex (str)(obj) = A single polygon vertex.
			select (bool) = Select the vertex after the operation. (default is True)
		'''
		pm.undoInfo(openChunk=True)
		connected_faces = pm.polyListComponentConversion(vertex, fromVertex=1, toFace=1) #pm.mel.PolySelectConvert(1) #convert to faces
		connected_faces_flat = pm.ls(connected_faces, flatten=1) #selectedFaces = pm.ls(sl=1, flatten=1)

		pm.polySplitVertex(vertex)

		#get a list for the vertices of each face that is connected to the original vertex.
		verts_sorted_by_face=[]
		for face in connected_faces_flat:

			connected_verts = pm.polyListComponentConversion(face, fromFace=1, toVertex=1) #pm.mel.PolySelectConvert(1) #convert to faces
			connected_verts_flat = [str(i) for i in pm.ls(connected_verts, flatten=1)] #selectedFaces = pm.ls(sl=1, flatten=1)
			verts_sorted_by_face.append(connected_verts_flat)

		out=[] #1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
		while len(verts_sorted_by_face)>0:
			first, rest = verts_sorted_by_face[0], verts_sorted_by_face[1:] #first, *rest = verts_sorted_by_face
			first = set(first)

			lf = -1
			while len(first)>lf:
				lf = len(first)

				rest2=[]
				for r in rest:
					if len(first.intersection(set(r)))>0:
						first |= set(r)
					else:
						rest2.append(r)     
				rest = rest2

			out.append(first)
			verts_sorted_by_face = rest


		for vertex_set in out:
			pm.polyMergeVertex(vertex_set, distance=0.001)

		pm.select(vertex_set, deselect=1) #deselect the vertices that were selected during the polyMergeVertex operation.
		if select:
			pm.select(vertex, add=1)
		pm.undoInfo(closeChunk=True)


	@staticmethod
	def getNGons(objects, repair=False):
		'''Get any N-Gons from the given object using selection contraints.

		:Parameters:
			objects (str)(obj)(list) = The objects to query.
			repair (bool) = Repair any found N-gons.

		:Return:
			(list)
		'''
		pm.select(objects)
		pm.mel.changeSelectMode(1) #Change to Component mode to retain object highlighting
		pm.selectType(smp=0, sme=1, smf=0, smu=0, pv=0, pe=1, pf=0, puv=0) #Change to Face Component Mode
		#Select Object/s and Run Script to highlight N-Gons
		pm.polySelectConstraint(mode=3, type=0x0008, size=3)
		nGons = pm.ls(sl=1)
		pm.polySelectConstraint(disable=1)

		if repair: #convert N-Sided Faces To Quads
			pm.polyQuad(nGons, angle=30, kgb=1, ktb=1, khe=1, ws=1)

		return nGons


	@staticmethod
	def getOverlappingVertices(objects, threshold=0.0003):
		'''Query the given objects for overlapping vertices.

		:Parameters:
			objects (str)(obj)(list) = The objects to query.
			threshold (float) = The maximum allowed distance.

		:Return:
			(list)
		'''
		import maya.OpenMaya as om
		from mayatk import mfnMeshGenerator

		result=[]
		for mfnMesh in mfnMeshGenerator(objects):
			points = om.MPointArray()
			mfnMesh.getPoints(points, om.MSpace.kWorld)

			for i in range(points.length()):
				for ii in range(points.length()):
					if i==ii:
						continue

					dist = points[i].distanceTo(points[ii])
					if dist < threshold:
						if i not in result:
							result.append(i)

						if ii not in result:
							result.append(ii)
		return result


	@classmethod
	def getOverlappingFaces(cls, objects):
		'''Get any duplicate overlapping faces of the given objects.

		::Parameters:
			objects (str)(obj)(list) = Faces or polygon objects.

		:Return:
			(list) duplicate overlapping faces.

		ex. call: pm.select(getOverlappingFaces(selection))
		'''
		if not objects:
			return []

		elif not pm.nodeType(objects)=='mesh': #if the objects are not faces.
			duplicates = itertk.flatten([cls.getOverlappingFaces(obj.faces) for obj in pm.ls(objects, objectsOnly=1)])
			return list(duplicates)

		face, *otherFaces = pm.ls(objects)
		face_vtx_positions = [v.getPosition() for v in pm.ls(pm.polyListComponentConversion(face, toVertex=1), flatten=1)]

		duplicates=[]
		for otherFace in otherFaces:
			otherFace_vtx_positions = [v.getPosition() for v in pm.ls(pm.polyListComponentConversion(otherFace, toVertex=1), flatten=1)]

			if face_vtx_positions==otherFace_vtx_positions: #duplicate found.
				duplicates.append(otherFace)
				otherFaces.remove(otherFace)

		if otherFaces:
			duplicates+=cls.getOverlappingFaces(otherFaces) #after adding any found duplicates, call again with any remaining faces.

		return duplicates


	@staticmethod
	def getSimilarMesh(obj, tol=0.0, includeOrig=False, **kwargs):
		'''Find similar geometry objects using the polyEvaluate command.
		Default behaviour is to compare all flags.

		:parameters:
			obj (str)(obj)(list) = The object to find similar for.
			tol (float) = The allowed difference in any of the given polyEvalute flag results (that return an int, float (or list of the int or float) value(s)).
			includeOrig (bool) = Include the original given obj with the return results.
			kwargs (bool) = Any keyword argument 'polyEvaluate' takes. Used to filter the results.
				ex: vertex, edge, face, uvcoord, triangle, shell, boundingBox, boundingBox2d, 
				vertexComponent, boundingBoxComponent, boundingBoxComponent2d, area, worldArea
		:return:
			(list) Similar objects.

		ex. call: getSimilarMesh(selection, vertex=1, area=1)
		'''
		lst = lambda x: list(x) if isinstance(x, (list, tuple, set)) else list(x.values()) if isinstance(x, dict) else [x] #assure the returned result from polyEvaluate is a list of values.

		obj, *other = pm.ls(obj, long=True, transforms=True)
		objProps = lst(pm.polyEvaluate(obj, **kwargs))

		otherSceneMeshes = set(pm.filterExpand(pm.ls(long=True, typ='transform'), selectionMask=12)) #polygon selection mask.
		similar = pm.ls([m for m in otherSceneMeshes if areSimilar(objProps, lst(pm.polyEvaluate(m, **kwargs)), tol=tol) and m!=obj])
		return similar+[obj] if includeOrig else similar


	@staticmethod
	def getSimilarTopo(obj, includeOrig=False, **kwargs):
		'''Find similar geometry objects using the polyCompare command.
		Default behaviour is to compare all flags.

		:parameters:
			obj (str)(obj)(list) = The object to find similar for.
			includeOrig (bool) = Include the original given obj with the return results.
			kwargs (bool) = Any keyword argument 'polyCompare' takes. Used to filter the results.
				ex: vertices, edges, faceDesc, uvSets, uvSetIndices, colorSets, colorSetIndices, userNormals
		:return:
			(list) Similar objects.
		'''
		obj, *other = pm.filterExpand(pm.ls(obj, long=True, tr=True), selectionMask=12) #polygon selection mask.

		otherSceneMeshes = set(pm.filterExpand(pm.ls(long=True, typ='transform'), sm=12))
		similar = pm.ls([m for m in otherSceneMeshes if pm.polyCompare(obj, m, **kwargs)==0 and m!=obj]) #0:equal,Verts:1,Edges:2,Faces:4,UVSets:8,UVIndices:16,ColorSets:32,ColorIndices:64,UserNormals=128. So a return value of 3 indicates both vertices and edges are different.
		return similar+[obj] if includeOrig else similar

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Parameters:
		attr (str): A name of a class attribute.

	:Return:
		(obj) The attribute.
	'''
	try:
		return getattr(Edit, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')







# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------


# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)

# -----------------------------------------------------------------------------






# -----------------------------------------------------------------------------

if __name__=='__main__':

	appendPaths('O:/Cloud/Code/_scripts/tentacle', verbose=False)

	from tentacle import import_submodules
	import test
	import_submodules('test')

	unittest.main(exit=False)



# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
import sys, os

import importlib
import inspect

try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)


class Script():
	'''
	'''
	@staticmethod
	def getMelGlobals(keyword=None, caseSensitive=False):
		'''Get global MEL variables.

		:Parameters:
			keyword (str) = search string.

		:Return:
			(list)
		'''
		variables = [
			v for v in sorted(pm.mel.eval('env')) 
				if not keyword 
					or (v.count(keyword) if caseSensitive else v.lower().count(keyword.lower()))
		]

		return variables


	@staticmethod
	def convertMelToPy(mel, excludeFromInput=[], excludeFromOutput=['from pymel.all import *','s pm']):
		'''Convert a string representing mel code into a string representing python code.

		:Parameters:
			mel (str) = string containing mel code.
			excludeFromInput (list) (list) = of strings specifying series of chars to strip from the Input.
			excludeFromOutput (list) (list) = of strings specifying series of chars to strip from the Output.
		
		mel2PyStr Parameters:
			currentModule (str) = The name of the module that the hypothetical code is executing in. In most cases you will leave it at its default, the __main__ namespace.
			pymelNamespace (str) = The namespace into which pymel will be imported. the default is '', which means from pymel.all import *
			forceCompatibility (bool) = If True, the translator will attempt to use non-standard python types in order to produce python code which more exactly reproduces the behavior of the original mel file, but which will produce 'uglier' code. Use this option if you wish to produce the most reliable code without any manual cleanup.
			verbosity (int) = Set to non-zero for a lot of feedback.
		'''
		from pymel.tools import mel2py
		import re

		l = filter(None, re.split('[\n][;]', mel))

		python=[]
		for e in l:
			if not e in excludeFromInput:
				try:
					py = mel2py.mel2pyStr(e, pymelNamespace='pm')
					for i in excludeFromOutput:
						py = py.strip(i)
				except:
					py = e
				python.append(py)

		return ''.join(python)


	@staticmethod
	def commandHelp(command): #mel command help
		#:Parameters: command (str) = mel command
		command = ('help ' + command)
		modtext = (pm.mel.eval(command))
		outputscrollField (modtext, "command help", 1.0, 1.0) #text, window_title, width, height


	@staticmethod
	def keywordSearch (keyword): #keyword command search
		#:Parameters: keyword (str) = 
		keyword = ('help -list' + '"*' + keyword + '*"')
		array = sorted(pm.mel.eval(keyword))
		outputTextField(array, "keyword search")


	@staticmethod
	def queryRuntime (command): #query runtime command info
		type       = ('whatIs '                           + command + ';')
		catagory   = ('runTimeCommand -query -category '  + command + ';')
		command	   = ('runTimeCommand -query -command '   + command + ';')
		annotation = ('runTimeCommand -query -annotation '+ command + ';')
		type = (pm.mel.eval(type))
		catagory = (pm.mel.eval(catagory))
		command = (pm.mel.eval(command))
		annotation = (pm.mel.eval(annotation))
		output_text = '{}\n\n{}\n{}\n\n{}\n{}\n\n{}\n{}\n\n{}\n{}'.format(command, "Type:", type, "Annotation:", annotation, "Catagory:", catagory, "Command:", command)
		outputscrollField(output_text, "runTimeCommand", 1.0, 1.0) #text, window_title, width, height


	@staticmethod
	def searchMEL (keyword): #search autodest MEL documentation
		url = '{}{}{}'.format("http://help.autodesk.com/cloudhelp/2017/ENU/Maya-Tech-Docs/Commands/",keyword,".html")
		pm.showHelp (url, absolute=True)


	@staticmethod
	def searchPython (keyword): #Search autodesk Python documentation
		url = '{}{}{}'.format("http://help.autodesk.com/cloudhelp/2017/ENU/Maya-Tech-Docs/CommandsPython/",keyword,".html")
		pm.showHelp (url, absolute=True)


	@staticmethod
	def searchPymel (keyword): #search online pymel documentation
		url = '{}{}{}'.format("http://download.autodesk.com/global/docs/maya2014/zh_cn/PyMel/search.html?q=",keyword,"&check_keywords=yes&area=default")
		pm.showHelp (url, absolute=True)


	@staticmethod
	def currentCtx(): #get current tool context
		output_text = pm.mel.eval('currentCtx;')
		outputscrollField(output_text, "currentCtx", 1.0, 1.0)


	@staticmethod
	def sourceScript(): #Source External Script file
		import os.path

		mel_checkBox = checkBox('mel_checkBox', query=1, value=1)
		python_checkBox = checkBox('python_checkBox', query=1, value=1)

		if mel_checkBox == 1:
			path = os.path.expandvars("%\CLOUD%/____graphics/Maya/scripts/*.mel")
			
		else:
			path = os.path.expandvars("%\CLOUD%/____graphics/Maya/scripts/*.py")

		file = pm.system.fileDialog (directoryMask=path)
		pm.openFile(file)


	@staticmethod
	def commandRef(): #open maya MEL commands list 
		pm.showHelp ('http://download.autodesk.com/us/maya/2011help/Commands/index.html', absolute=True)


	@staticmethod
	def globalVars(): #$g List all global mel variables in current scene
		pm.mel.eval('scriptEditorInfo -clearHistory')
		array = sorted(pm.mel.eval("env"))
		outputTextField(array, "Global Variables")


	@staticmethod
	def listUiObjects(): #lsUI returns the names of UI objects
		windows			= '{}\n{}\n{}\n'.format("Windows", "Windows created using ELF UI commands:", pm.lsUI (windows=True))
		panels			= '{}\n{}\n{}\n'.format("Panels", "All currently existing panels:", pm.lsUI (panels=True))
		editors			= '{}\n{}\n{}\n'.format("Editors", "All currently existing editors:", pm.lsUI (editors=True))
		controls		= '{}\n{}\n{}\n'.format("Controls", "Controls created using ELF UI commands: [e.g. buttons, checkboxes, etc]", pm.lsUI (controls=True))
		control_layouts = '{}\n{}\n{}\n'.format("Control Layouts", "Control layouts created using ELF UI commands: [e.g. formLayouts, paneLayouts, etc.]", pm.lsUI (controlLayouts=True))
		menus			= '{}\n{}\n{}\n'.format("Menus", "Menus created using ELF UI commands:", pm.lsUI (menus=True))
		menu_items	= '{}\n{}\n{}\n'.format("Menu Items", "Menu items created using ELF UI commands:", pm.lsUI (menuItems=True))
		contexts		= '{}\n{}\n{}\n'.format("Tool Contexts", "Tool contexts created using ELF UI commands:", pm.lsUI (contexts=True))
		output_text	= '{}\n{}\n{}\n{}\n{}\n{}\n{}\n{}'.format(windows, panels, editors, menus, menu_items, controls, control_layouts, contexts)
		outputscrollField(output_text, "Ui Elements", 6.4, 0.85)

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Return:
		(obj)
	'''
	try:
		return getattr(Script, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')




# print (__package__, __file__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect


# ------------------------------------------------------------------------------------
# this is the missing stuff when running python.exe compared with mayapy.exe

# mayaver = 2022
# pythonver = 37

# mayapath = '%ProgramFiles%/Autodesk/Maya{}'.format(mayaver)

# os.environ['MAYA_LOCATION'] = mayapath
# os.environ['PYTHONHOME'] = mayapath+'/Python{}'.format(mayaver, pythonver)
# os.environ['PATH'] = mayapath+'/bin;'.format(mayaver) + os.environ['PATH']

# from pythontk import File
# for d in [
# 	'{}/bin'.format(mayapath), 
# 	'{}/bin3'.format(mayapath), 
# 	'{}/Python{}'.format(mayapath, pythonver)
# 	]:
# 	for dd in File.getDirContents(d, 'dirpaths', excDirs='Python27',  recursive=True):
# 		print (dd)
# 		sys.path.append(dd)

# import maya.standalone
# maya.standalone.initialize(name='python')

import pymel.core as pm

from mayatk import *
import test


sfr = pm.melGlobals['cmdScrollFieldReporter']
pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)


class Mayatk_test(Main, Mayatk):
	'''
	set object mode:
		pm.selectMode(object=1)

	set component mode:
		pm.selectMode(component=1)

	set component mode type:
		pm.selectType(allObjects=1)
		pm.selectType(mc=1)
		pm.selectType(vertex=1)
		pm.selectType(edge=1)
		pm.selectType(facet=1)
		pm.selectType(polymeshUV=1)
		pm.selectType(meshUVShell=1)
	'''
	#Tear down the any previous test by creating a new scene:
	pm.mel.file(new=True, force=True)

	#assemble the test scene:
	if not pm.objExists('cyl'):
		cyl = pm.polyCylinder(radius=5, height=10, subdivisionsX=12, subdivisionsY=1, subdivisionsZ=1, name='cyl')

	def test_undo(self):
		'''
		'''
		self.perform_test({
			"str(self.undo()).split(' at')[0]": '<function Mayatk.undo.<locals>.wrapper',
		})

	def test_getMainWindow(self):
		'''
		'''
		self.perform_test({
			"self.getMainWindow()": None,
		})

	def test_getType(self):
		'''
		'''
		self.perform_test({
			"self.getType('cyl')": 'transform',
			"self.getType('cylShape')": 'mesh',
			"self.getType('cylShape.vtx[0]')": 'vtx',
			"self.getType('cylShape.e[0]')": 'e',
			"self.getType('cylShape.f[0]')": 'f',
		})

	def test_getTransformNode(self):
		'''
		'''
		self.perform_test({
			"self.getTransformNode('cyl')": 'cyl',
			"self.getTransformNode('cylShape')": 'cyl',
		})

	def test_getShapeNode(self):
		'''
		'''
		self.perform_test({
			"self.getShapeNode('cyl')": 'cylShape',
			"self.getShapeNode('cylShape')": 'cylShape',
		})

	def test_getHistoryNode(self):
		'''
		'''
		self.perform_test({
			"self.getHistoryNode('cyl')": 'polyCylinder1',
			"self.getHistoryNode('cylShape')": 'polyCylinder1',
		})

	def test_mfnMeshGenerator(self):
		'''
		'''
		self.perform_test({
			"str(next(self.mfnMeshGenerator('cyl'))).split(';')[0]": '<maya.OpenMaya.MFnMesh',
		})

	def test_isLocator(self):
		'''
		'''
		if not pm.objExists('loc'):
			loc = pm.spaceLocator(name='loc')

		self.perform_test({
			"self.isLocator('cyl')": False,
			"self.isLocator('loc')": True,
		})

	def test_isGroup(self):
		'''
		'''
		self.perform_test({
			"self.isGroup('cyl')": False,
			"self.isGroup('cylShape')": False,
			"self.isGroup('cylShape.vtx[0]')": False,
		})

	def test_getGroups(self):
		'''
		'''
		self.perform_test({
			"self.getGroups()": [],
		})

	def test_getParent(self):
		'''
		'''
		self.perform_test({
			"self.getParent('cyl')": None,
		})

	def test_getChildren(self):
		'''
		'''
		self.perform_test({
			"self.getChildren('cyl')": [],
		})

	def test_getArrayType(self):
		'''
		'''
		self.perform_test({
			"self.getArrayType(100)": 'int',
			"self.getArrayType('cylShape.vtx[:]')": 'str',
			"self.getArrayType(pm.ls('cylShape.vtx[:]'))": 'obj',
		})

	def test_convertArrayType(self):
		'''
		'''
		self.perform_test({
			"self.convertArrayType('cyl.vtx[:2]', 'str')": ['cylShape.vtx[0:2]'],
			"self.convertArrayType('cyl.vtx[:2]', 'str', flatten=True)": ['cylShape.vtx[0]', 'cylShape.vtx[1]', 'cylShape.vtx[2]'],
			"str(self.convertArrayType('cyl.vtx[:2]', 'obj'))": "[MeshVertex('cylShape.vtx[0:2]')]",
			"str(self.convertArrayType('cyl.vtx[:2]', 'obj', flatten=True))": "[MeshVertex('cylShape.vtx[0]'), MeshVertex('cylShape.vtx[1]'), MeshVertex('cylShape.vtx[2]')]",
			"self.convertArrayType('cyl.vtx[:2]', 'int')": [0, 2],
			"self.convertArrayType('cyl.vtx[:2]', 'int', flatten=True)": [0, 1, 2],
		})

	def test_getParameterValuesMEL(self):
		'''
		'''
		self.perform_test({
			# "self.getParameterValuesMEL()": None,
		})

	def test_setParameterValuesMEL(self):
		'''
		'''
		self.perform_test({
			# "self.setParameterValuesMEL()": None,
		})

	def test_getAttributesMEL(self):
		'''
		'''
		self.perform_test({
			# "self.getAttributesMEL()": None,
		})

	def test_setAttributesMEL(self):
		'''
		'''
		self.perform_test({
			# "self.setAttributesMEL()": None,
		})

	def test_connectAttributes(self):
		'''
		'''
		self.perform_test({
			# "self.connectAttributes()": None,
		})

	def test_createRenderNode(self):
		'''
		'''
		self.perform_test({
			# "self.createRenderNode()": None,
		})

	def test_getIncomingNodeByType(self):
		'''
		'''
		self.perform_test({
			# "self.getIncomingNodeByType()": None,
		})

	def test_getOutgoingNodeByType(self):
		'''
		'''
		self.perform_test({
			# "self.getOutgoingNodeByType()": None,
		})

	def test_connectMultiAttr(self):
		'''
		'''
		self.perform_test({
			# "self.connectMultiAttr()": None,
		})

	def test_nodeExists(self):
		'''
		'''
		self.perform_test({
			# "self.nodeExists()": None,
		})

	def test_getSelectedChannels(self):
		'''
		'''
		self.perform_test({
			# "self.getSelectedChannels()": None,
		})

	def test_getPanel(self):
		'''
		'''
		self.perform_test({
			# "self.getPanel()": None,
		})

	def test_mainProgressBar(self):
		'''
		'''
		self.perform_test({
			# "self.mainProgressBar()": None,
		})

	def test_viewportMessage(self):
		'''
		'''
		self.perform_test({
			# "self.viewportMessage()": None,
		})

# -----------------------------------------------------------------------------

if __name__=='__main__':

	unittest.main(exit=False)






# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		"self.()": None,
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})
# """

# Deprecated ---------------------


# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from pythontk import Iter, Math, randomize


class _GetComponents():
	'''
	'''
	componentTypes = [
		#abv 	singular 	plural 		full						int		hex 
		('vtx', 'vertex', 	'vertices', 'Polygon Vertex', 			31, 	0x0001),
		('e', 	'edge', 	'edges', 	'Polygon Edge', 			32,		0x8000),
		('f', 	'face', 	'faces', 	'Polygon Face', 			34, 	0x0008),
		('uv', 	'texture', 	'texture coordinates', 'Polygon UV', 	35, 	0x0010),
		('cv', 	'control vertex', 'control vertices', 'Control Vertex', 28, None),
		('vtxf', 'vertexFace', 'vertexFaces', 'Polygon Vertex Face', 70, 	None),
		(None, 	None, 		None, 		'Edit Point', 				30, 	None),
		(None, 	None, 		None, 		'Handle', 					0, 		None),
		(None, 	None, 		None, 		'Nurbs Curves On Surface', 	11, 	None),
		(None, 	None, 		None, 		'Subdivision Mesh Point', 	36, 	None),
		(None, 	None, 		None, 		'Subdivision Mesh Edge', 	37, 	None), 
		(None, 	None, 		None, 		'Subdivision Mesh Face', 	38, 	None), 
		(None, 	None, 		None, 		'Curve Parameter Point', 	39, 	None), 
		(None, 	None, 		None, 		'Curve Knot', 				40, 	None),
		(None, 	None, 		None, 		'Surface Parameter Point', 	41, 	None),
		(None, 	None, 		None, 		'Surface Knot', 			42, 	None),
		(None, 	None, 		None, 		'Surface Range', 			43, 	None),
		(None, 	None, 		None, 		'Trim Surface Edge', 		44, 	None),
		(None, 	None, 		None, 		'Surface Isoparm', 			45, 	None),
		(None, 	None, 		None, 		'Lattice Point', 			46, 	None),
		(None, 	None, 		None, 		'Particle', 				47, 	None),
		(None, 	None, 		None, 		'Scale Pivot', 				49, 	None),
		(None, 	None, 		None, 		'Rotate Pivot', 			50, 	None),
		(None, 	None, 		None, 		'Select Handle', 			51, 	None),
		(None, 	None, 		None, 		'NURBS Surface Face', 		72, 	None),
		(None, 	None, 		None, 		'Subdivision Mesh UV', 		73, 	None),
	]

	@classmethod
	def getComponentType(cls, component, nom='abv'):
		'''Get the type of a given component.

		:Parameters:
			obj (str)(obj)(list) = A single maya component.
				If multiple components are given, only the first will be sampled.
			nom (str) = Specify the desired return value type. (default: 'str')
				(valid: 'full' - object type as a string.
						'int' - maya mask value as an integer.
						'hex' - hex value. ie. 0x0001
						'abv' - abreviated object type as a string. ie. 'vtx'
		:Return:
			(str)(int) dependant on 'nom' arg.

		ex. call:
		getComponentType('cyl.e[:]') #returns: 'e'
		getComponentType('cyl.vtx[:]', 'abv') #returns: 'vtx'
		getComponentType('cyl.e[:]', 'int') #returns: 32
		'''
		for a, s, p, f, i, h in cls.componentTypes:
			try:
				if pm.filterExpand(component, sm=i):
					if nom=='abv': return a
					elif nom=='full': return f
					elif nom=='int': return i
					elif nom=='hex': return h
					elif nom=='plural': return p
					else: return s
			except Exception as error:
				print ('File "{}" in getComponentType\n# Error: Not a valid component. #\n	{}{}'.format(__file__, error, '(empty string)' if component=='' else ''))
				break
		return None


	@classmethod
	def convertAlias(cls, componentType, nom='abv'):
		'''Return an alternate component alias for the given alias. 
		ie. a hex value of 0x0001 for 'vertex'
		If nothing is found, a value of 'None' will be returned.

		:Parameters:
			componentType () = A component type. ex. 'vertex', 'vtx', 31, or 0x0001
			nom (str) = The desired returned alias.  (default: 'abv')
				(valid: 'abv', 'singular', 'plural', 'str', 'int', 'hex')

		:Return:
			(str)(int)(hex)(None) dependant on 'nom' argument.

		ex. call:
		convertAlias('vertex', 'hex') #returns: 0x0001
		convertAlias(0x0001, 'str') #returns: 'Polygon Vertex'
		'''
		rtypes = ('abv', 'singular', 'plural', 'full', 'int', 'hex')

		for t in cls.componentTypes:
			if componentType in t:
				index = rtypes.index(nom)
				return t[index]
		return None


	@classmethod
	def convertComponentType(cls, components, componentType, returnType='str', flatten=False):
		'''Convert component(s) to it's sub-components of the given type.

		:Parameters:
			components (str)(obj)(list) = The components(s) to convert.
			componentType (str) = The desired returned component type. 
				valid: 'vtx' (or 'vertex', 'vertices', 'Polygon Vertex', 31, 0x0001), 
					and the same for each: 'edge', 'uv', 'face'.
			returnType (str) = The desired returned object type. 
				(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list)(dict)

		ex. call:
		convertComponentType('obj.vtx[:2]', 'vertex') #returns: ['obj.vtx[0:2]']
		convertComponentType('obj.vtx[:2]', 'face') #returns: ['obj.f[0:2]', 'obj.f[11:14]', 'obj.f[23]']
		convertComponentType('obj.vtx[:2]', 'edge') #returns: ['obj.e[0:2]', 'obj.e[11]', 'obj.e[24:26]', 'obj.e[36:38]']
		convertComponentType('obj.vtx[:2]', 'uv') #returns: ['obj.map[0:2]', 'obj.map[12:14]', 'obj.map[24]']
		'''
		d = {'vtx':'toVertex', 'e':'toEdge', 'uv':'toUV', 'f':'toFace', 'uv':'toUV', 'shell':'toShell', 'vertexFace':'toVertexFace'}
		typ = cls.convertAlias(componentType) #get the correct componentType variable from possible args.

		if not typ in d:
			return components
		components = pm.polyListComponentConversion(components, **{d[typ.lower()]:True})
		return convertArrayType(components, returnType=returnType, flatten=flatten)


	@classmethod
	def convertIntToComponent(cls, obj, integers, componentType, returnType='str', flatten=False):
		'''Convert the given integers to components of the given object.

		:Parameters:
			obj (str)(obj)(list) = The object to convert to vertices of.
			integers (list) = The integer(s) to convert.
			componentType (str) = The desired returned component type. 
				valid: 'vtx' (or 'vertex', 'vertices', 'Polygon Vertex', 31, 0x0001), 
					and the same for each: 'edge', 'uv', 'face'.
			returnType (str) = The desired returned object type. 
				(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list)

		ex. call: convertIntToComponent('cyl', range(4), 'f') #returns: ['cylShape.f[0:3]']
		'''
		obj = pm.ls(obj, objectsOnly=True)[0]
		objName = obj.name()

		if not flatten:
			n = lambda c: '{}:{}'.format(c[0], c[-1]) if len(c)>1 else str(c[0])
			result = ['{}.{}[{}]'.format(objName, componentType, n(c)) for c in Iter.splitList(integers, 'range')]
		else:
			result = ['{}.{}[{}]'.format(objName, componentType, c) for c in integers]

		return convertArrayType(result, returnType=returnType, flatten=flatten)


	@classmethod
	def filterComponents(cls, components, inc=[], exc=[], flatten=False):
		'''Filter the given components.

		:Parameters:
			components (str)(obj)(list) = The components(s) to filter.
			inc (str)(int)(obj)(list) = The component(s) to include.
			exc (str)(int)(obj)(list) = The component(s) to exclude.
						(exlude take precidence over include)
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list)

		ex. call:
		filterComponents('cyl.vtx[:]', 'cyl.vtx[:2]', 'cyl.vtx[1:23]') #returns: ['cyl.vtx[0]']
		filterComponents('cyl.f[:]', range(2), range(1, 23)) #returns: ['cyl.f[0]']
		'''
		from mayatk import getArrayType, convertArrayType

		typ = cls.getComponentType(components)
		etyp = getArrayType(components)
		etyp_inc = getArrayType(inc)
		etyp_exc = getArrayType(exc)

		if etyp_inc=='int' or etyp_exc=='int':
			try:
				obj = pm.ls(components, objectsOnly=True)[0]
			except IndexError as error:
				print ('File "{}" in filterComponents\n# Error: Operation requires at least one component. #\n	{}'.format(__file__, error))
				return []

		if etyp_inc=='int':
			inc = cls.convertIntToComponent(obj, inc, typ)
		inc = pm.ls(inc, flatten=True)

		if etyp_exc=='int':
			exc = cls.convertIntToComponent(obj, exc, typ)
		exc = pm.ls(exc, flatten=True)

		components = pm.ls(components, flatten=True)

		filtered = Iter.filterList(components, inc=inc, exc=exc)
		result = convertArrayType(filtered, returnType=etyp, flatten=flatten)
		return result


	@classmethod
	def getComponents(cls, objects, componentType, returnType='str', inc=[], exc=[], randomize=0, flatten=False):
		'''Get the components of the given type from the given object(s).
		If no objects are given the current selection will be used.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to get the components of. (Polygon, Polygon components)(default: current selection)
			componentType (str)(int) = The component type to return. (valid: any type allowed in the 'convertAlias' method)
			returnType (str) = The desired returned object type.
				(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
			inc (str)(obj)(list) = The component(s) to include.
			exc (str)(obj)(list) = The component(s) to exclude. (exlude take precidence over include)
			randomize (float) = If a 0.1-1 value is given, random components will be returned with a quantity determined by the given ratio. 
								A value of 0.5 will return a 50% of the components of an object in random order.
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list)(dict) Dependant on flags.

		ex. call:
		getComponents('obj', 'vertex', 'str', '', 'obj.vtx[2:23]') #returns: ['objShape.vtx[0]', 'objShape.vtx[1]', 'objShape.vtx[24]', 'objShape.vtx[25]']
		getComponents('obj', 'vertex', 'obj', '', 'obj.vtx[:23]') #returns: [MeshVertex('objShape.vtx[24]'), MeshVertex('objShape.vtx[25]')]
		getComponents('obj', 'f', 'int') #returns: {nt.Mesh('objShape'): [(0, 35)]}
		getComponents('obj', 'edges') #returns: ['objShape.e[0:59]']
		getComponents('obj', 'edges', 'str', 'obj.e[:2]') #returns: ['objShape.e[0]', 'objShape.e[1]', 'objShape.e[2]']
		'''

		components = cls.convertComponentType(objects, componentType)

		if inc or exc:
			components = cls.filterComponents(components, inc=inc, exc=exc)

		if randomize:
			components = randomize(pm.ls(components, flatten=1), randomize)

		result = convertArrayType(components, returnType=returnType, flatten=flatten)
		return result



class Cmpt(_GetComponents):
	'''
	'''
	@classmethod
	def getContigiousEdges(cls, components):
		'''Get a list containing sets of adjacent edges.

		:Parameters:
			components (list) = Polygon components to be filtered for adjacent edges.

		:Return:
			(list) adjacent edge sets.

		ex. call:
		getContigiousEdges(['obj.e[:2]']) #returns: [{'objShape.e[1]', 'objShape.e[0]', 'objShape.e[2]'}]
		getContigiousEdges(['obj.f[0]']) #returns: [{'objShape.e[24]', 'objShape.e[0]', 'objShape.e[25]', 'objShape.e[12]'}]
		'''
		edges = cls.convertComponentType(components, 'edge', flatten=1)

		sets=[]
		for edge in edges:
			vertices = pm.polyListComponentConversion(edge, fromEdge=1, toVertex=1)
			connEdges = cls.convertComponentType(vertices, 'edge', flatten=1)
			edge_set = set([e for e in connEdges if e in edges]) #restrict the connected edges to the original edge pool.
			sets.append(edge_set)

		result=[]
		while len(sets)>0: #combine sets in 'sets' that share common elements.
			first, rest = sets[0], sets[1:] #python 3: first, *rest = sets
			first = set(first)

			lf = -1
			while len(first)>lf:
				lf = len(first)

				rest2 = []
				for r in rest:
					if len(first.intersection(set(r)))>0:
						first |= set(r)
					else:
						rest2.append(r)     
				rest = rest2

			result.append(first)
			sets = rest

		return result


	@classmethod
	def getContigiousIslands(cls, faces, faceIslands=[]):
		'''Get a list containing sets of adjacent polygon faces grouped by islands.

		:Parameters:
			faces (str)(obj)(list) = The polygon faces to be filtered for adjacent.
			faceIslands (list) = optional. list of sets. ability to add faces from previous calls to the return value.

		:Return:
			(list) of sets of adjacent faces.

		ex. call: getContigiousIslands('obj.f[21:26]') #returns: [{'objShape.f[22]', 'objShape.f[21]', 'objShape.f[23]'}, {'objShape.f[26]', 'objShape.f[24]', 'objShape.f[25]'}]
		'''
		sets=[]
		faces = pm.ls(faces, flatten=1)
		for face in faces:
			edges = pm.polyListComponentConversion(face, fromFace=1, toEdge=1)
			borderFaces = cls.convertComponentType(edges, 'face', 'obj', flatten=1)
			set_ = set([str(f) for f in borderFaces if f in faces])
			if set_:
				sets.append(set_)

		while len(sets)>0: #combine sets in 'sets' that share common elements.
			first, rest = sets[0], sets[1:] #python 3: first, *rest = sets
			first = set(first)

			lf = -1
			while len(first)>lf:
				lf = len(first)

				rest2 = []
				for r in rest:
					if len(first.intersection(set(r)))>0:
						first |= set(r)
					else:
						rest2.append(r)     
				rest = rest2

			faceIslands.append(first)
			sets = rest

		return faceIslands


	@staticmethod
	def getIslands(obj, returnType='str', flatten=False):
		'''Get the group of components in each separate island of a combined mesh.

		:parameters:
			obj (str)(obj)(list) = The object to get shells from.
			returnType (bool) = Return the shell faces as a list of type: 'str' (default), 'int', or 'obj'.
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:return:
			(generator)

		ex. call: getIslands('combined_obj') #returns: [['combined_obj.f[0]', 'combined_obj.f[5]', ..etc, ['combined_obj.f[15]', ..etc]] 
		'''
		num_shells = pm.polyEvaluate(obj, shell=True)
		num_faces = pm.polyEvaluate(obj, face=True)

		unprocessed = set(range(num_faces))

		shells=[]
		while unprocessed:
			index = next(iter(unprocessed)) #face_index
			faces = pm.polySelect(obj, extendToShell=index, noSelection=True) #shell faces

			if returnType=='str':
				yield ["{}.f[{}]".format(obj, index) for index in faces]

			elif returnType=='int':
				yield [index for index in faces]

			elif returnType=='obj':
				yield [pm.ls("{}.f[{}]".format(obj, index))[0] for index in faces]

			unprocessed.difference_update(faces)


	@classmethod
	def getBorderComponents(cls, x, componentType='', returnType='str', componentBorder=False, flatten=False):
		'''Get any object border components from given component(s) or a polygon object.
		A border is defined as a hole or detached edge.

		:Parameters:
			x (str)(obj)(list) = Component(s) (or a polygon object) to find any border components for.
			componentType (str) = The desired returned component type. (valid: 'vertex','edge','face', '',
				An empty string returns the same type as the first given component, or edges if an object is given)
			returnType (str) = The desired returned object type.
				(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
			componentBorder (bool) = Get the components that border given components instead of the mesh border.
				(valid: 'component', 'object'(default))
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list) components that border an open edge.

		ex. call:
		getBorderComponents('pln', 'vtx') #returns: ['plnShape.vtx[0:4]', 'plnShape.vtx[7:8]', 'plnShape.vtx[11:15]'],
		getBorderComponents('pln') #returns: ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
		getBorderComponents('pln.e[:]') #returns: ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
		getBorderComponents(['pln.e[9]','pln.e[10]', 'pln.e[12]', 'pln.e[16]'], 'f', componentBorder=True) #returns: ['plnShape.f[1]', 'plnShape.f[3:5]', 'plnShape.f[7]'],
		getBorderComponents('pln.f[3:4]', 'vtx', componentBorder=True) #returns: ['plnShape.vtx[4:6]', 'plnShape.vtx[8:10]'],
		'''
		if not x:
			print ('File "{}" in getBorderComponents\n# Error: Operation requires a given object(s) or component(s). #'.format(__file__))
			return []

		origType = cls.getComponentType(x, 'abv')
		if not origType:
			origType, x = 'mesh', cls.getComponents(x, 'edges')
		origVerts = cls.convertComponentType(x, 'vtx', flatten=True)
		origEdges = cls.convertComponentType(x, 'edge', flatten=True)
		origFaces = cls.convertComponentType(x, 'face', flatten=True)

		if not componentType: #if no component type is specified, return the same type of component as given. in the case of mesh object, edges will be returned.
			componentType = origType if not origType=='mesh' else 'e'
		else:
			componentType = cls.convertAlias(componentType) #get the correct componentType variable from possible args.

		result=[]
		if componentBorder: #get edges Qthat form the border of the given components.
			for edge in origEdges:
				attachedFaces = cls.convertComponentType(edge, 'face', flatten=1)
				if componentType=='f':
					for f in attachedFaces:
						if origType=='f' and f in origFaces:
							continue
						result.append(f)
					continue
				attachedEdges = cls.convertComponentType(attachedFaces, 'edge', flatten=1)
				for e in attachedEdges:
					if origType=='e' and e in origEdges:
						continue
					attachedVerts = cls.convertComponentType(e, 'vtx', flatten=1)
					for v in attachedVerts:
						if v in origVerts:
							result.append(v)

		else: #get edges that form the border of the object.
			for edge in origEdges:
				attachedFaces = cls.convertComponentType(edge, 'face', flatten=1)
				if len(attachedFaces)==1:
					result.append(edge)

		result = cls.convertComponentType(result, componentType) #convert back to the original component type and flatten /un-flatten list.
		result = convertArrayType(result, returnType=returnType, flatten=flatten)
		return result


	@classmethod
	def getClosestVerts(cls, set1, set2, tolerance=1000):
		'''Find the two closest vertices between the two sets of vertices.

		:Parameters:
			set1 (str)(list) = The first set of vertices.
			set2 (str)(list) = The second set of vertices.
			tolerance (float) = Maximum search distance.

		:Return:
			(list) closest vertex pairs by order of distance (excluding those not meeting the tolerance). (<vertex from set1>, <vertex from set2>).

		ex. call: getClosestVerts('pln.vtx[:10]', 'pln.vtx[11:]', 6.667) #returns: [('plnShape.vtx[7]', 'plnShape.vtx[11]'), ('plnShape.vtx[8]', 'plnShape.vtx[12]'), ('plnShape.vtx[9]', 'plnShape.vtx[13]'), ('plnShape.vtx[10]', 'plnShape.vtx[11]'), ('plnShape.vtx[10]', 'plnShape.vtx[14]')]
		'''
		from operator import itemgetter

		set1 = convertArrayType(set1, returnType='str', flatten=True)
		set2 = convertArrayType(set2, returnType='str', flatten=True)
		vertPairsAndDistance={}
		for v1 in set1:
			v1Pos = pm.pointPosition(v1, world=1)
			for v2 in set2:
				v2Pos = pm.pointPosition(v2, world=1)
				distance = Math.getDistBetweenTwoPoints(v1Pos, v2Pos)
				if distance<tolerance:
					vertPairsAndDistance[(v1, v2)] = distance

		sorted_ = sorted(vertPairsAndDistance.items(), key=itemgetter(1))
		vertPairs = [i[0] for i in sorted_]

		return vertPairs


	@classmethod
	def getClosestVertex(cls, vertices, obj, tolerance=0.0, freezeTransforms=False, returnType='str'):
		'''Find the closest vertex of the given object for each vertex in the list of given vertices.

		:Parameters:
			vertices (list) = A set of vertices.
			obj (str)(obj)(list) = The reference object in which to find the closest vertex for each vertex in the list of given vertices.
			tolerance (float) = Maximum search distance. Default is 0.0, which turns off the tolerance flag.
			freezeTransforms (bool) = Reset the selected transform and all of its children down to the shape level.
			returnType (str) = The desired returned object type. This only affects the dict value (found vertex), 
					the key (orig vertex) is always a string. ex. {'planeShape.vtx[0]': 'objShape.vtx[3]'} vs. {'planeShape.vtx[0]': MeshVertex('objShape.vtx[3]')}
					(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
		:Return:
			(dict) closest vertex pairs {<vertex from set1>:<vertex from set2>}.

		ex. call:
		getClosestVertex('plnShape.vtx[0]', 'cyl', returnType='int') #returns: {'plnShape.vtx[0]': 3},
		getClosestVertex('plnShape.vtx[0]', 'cyl') #returns: {'plnShape.vtx[0]': 'cylShape.vtx[3]'},
		getClosestVertex('plnShape.vtx[2:3]', 'cyl') #returns: {'plnShape.vtx[2]': 'cylShape.vtx[2]', 'plnShape.vtx[3]': 'cylShape.vtx[1]'}
		'''
		vertices = convertArrayType(vertices, returnType='str', flatten=True)
		pm.undoInfo(openChunk=True)

		if freezeTransforms:
			pm.makeIdentity(obj, apply=True)

		obj2Shape = pm.listRelatives(obj, children=1, shapes=1)[0] #pm.listRelatives(obj, fullPath=False, shapes=True, noIntermediate=True)

		cpmNode = pm.ls(pm.createNode('closestPointOnMesh'))[0] #create a closestPointOnMesh node.
		pm.connectAttr(obj2Shape.outMesh, cpmNode.inMesh, force=1) #object's shape mesh output to the cpm node.

		closestVerts={}
		for v1 in vertices: #assure the list of vertices is a flattened list of stings. prevent unhashable type error when closestVerts[v1] = v2.  may not be needed with python versions 3.8+
			v1Pos = pm.pointPosition(v1, world=True)
			pm.setAttr(cpmNode.inPosition, v1Pos[0], v1Pos[1], v1Pos[2], type="double3") #set a compound attribute

			index = pm.getAttr(cpmNode.closestVertexIndex) #vertex Index. | ie. result: [34]
			v2 = obj2Shape.vtx[index]

			v2Pos = pm.pointPosition(v2, world=True)
			distance = Math.getDistBetweenTwoPoints(v1Pos, v2Pos)

			v2_convertedType = convertArrayType(v2, returnType=returnType)[0]
			if not tolerance:
				closestVerts[v1] = v2_convertedType
			elif distance < tolerance:
				closestVerts[v1] = v2_convertedType

		pm.delete(cpmNode)
		pm.undoInfo(closeChunk=True)

		return closestVerts


	@classmethod
	def getEdgePath(cls, components, path='edgeLoop', returnType='str', flatten=False):
		'''Query the polySelect command for the components along different edge paths.
		Supports components from a single object.

		:Parameters:
			components (str)(obj)(list) = The components used for the query (dependant on the operation type).
			path (str) = The desired return type. valid: 'edgeLoop': Select an edge loop starting at the given edge.
				'edgeRing': Select an edge ring starting at the given edge.
				'edgeRingPath', Given two edges that are on the same edge ring, this will select the shortest path between them on the ring.
				'edgeLoopPath': Given two edges that are on the same edge loop, this will select the shortest path between them on the loop.
			returnType (str) = The desired returned object type.
				(valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list) The components comprising the path.

		ex. call:
		getEdgePath('sph.e[12]', 'edgeLoop') #returns: ['sphShape.e[12]', 'sphShape.e[17]', 'sphShape.e[16]', 'sphShape.e[15]', 'sphShape.e[14]', 'sphShape.e[13]']
		getEdgePath('sph.e[12]', 'edgeLoop', 'int') #returns: [12, 17, 16, 15, 14, 13]
		getEdgePath('sph.e[12]', 'edgeRing') #returns: ['sphShape.e[0]', 'sphShape.e[6]', 'sphShape.e[12]', 'sphShape.e[18]', 'sphShape.e[24]']
		getEdgePath(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath') #returns: ['sphShape.e[43]', 'sphShape.e[42]', 'sphShape.e[47]', 'sphShape.e[46]']
		getEdgePath(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath') #returns: ['sphShape.e[60]', 'sphShape.e[48]', 'sphShape.e[42]', 'sphShape.e[36]', 'sphShape.e[30]', 'sphShape.e[54]']
		'''
		obj, *other = pm.ls(components, objectsOnly=1)
		cnums = cls.convertComponentType(components, 'edge', returnType='int', flatten=True)

		if len(cnums)<2 and path in ('edgeRingPath', 'edgeLoopPath'):
			print ('File "{}" in getEdgePath\n# Error: Operation requires at least two components. #\n	Edges given: {}'.format(__file__, cnums))
			return []

		if path=='edgeRing':
			edgesLong = pm.polySelect(obj, query=1, edgeRing=cnums) #(e..)

		elif path=='edgeRingPath':
			edgesLong = pm.polySelect(obj, query=1, edgeRingPath=(cnums[0], cnums[1])) #(e, e)
			if not edgesLong:
				print ('File "{}" in getEdgePath\n# Error: getEdgePath: Operation requires two edges that are on the same edge ring. #\n	Edges given: {}, {}'.format(__file__, cnums[0], cnums[1]))
				return []

		elif path=='edgeLoopPath':
			edgesLong = pm.polySelect(obj, query=1, edgeLoopPath=(cnums[0], cnums[1])) #(e, e)
			if not edgesLong:
				print ('File "{}" in getEdgePath\n# Error: getEdgePath: Operation requires two edges that are on the same edge loop. #\n	Edges given: {}, {}'.format(__file__, cnums[0], cnums[1]))
				return []
		else: #'edgeLoop'
			edgesLong = pm.polySelect(obj, query=1, edgeLoop=cnums) #(e..)

		objName = obj.name()
		result = Iter.removeDuplicates(['{}.e[{}]'.format(objName, e) for e in edgesLong])
		return convertArrayType(result, returnType=returnType, flatten=flatten)


	@classmethod
	def getShortestPath(cls, components, returnType='str', flatten=False):
		'''Get the shortest path between two components.

		:Parameters:
			components (obj) = A Pair of vertices or edges.
			returnType (str) = The desired returned object type. 
				valid: 'str'(default), 'obj', 'int'(valid only at sub-object level)
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list) the components that comprise the path as strings.

		ex. call:
		getEdgePath('sph.e[12]', 'edgeLoop') #returns: ['sphShape.e[12]', 'sphShape.e[17]', 'sphShape.e[16]', 'sphShape.e[15]', 'sphShape.e[14]', 'sphShape.e[13]']
		getEdgePath('sph.e[12]', 'edgeLoop', 'int') #returns: [12, 17, 16, 15, 14, 13]
		getEdgePath('sph.e[12]', 'edgeRing') #returns: ['sphShape.e[0]', 'sphShape.e[6]', 'sphShape.e[12]', 'sphShape.e[18]', 'sphShape.e[24]']
		getEdgePath(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath') #returns: ['sphShape.e[43]', 'sphShape.e[42]', 'sphShape.e[47]', 'sphShape.e[46]']
		getEdgePath(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath') #returns: ['sphShape.e[60]', 'sphShape.e[48]', 'sphShape.e[42]', 'sphShape.e[36]', 'sphShape.e[30]', 'sphShape.e[54]']
		'''
		obj = pm.ls(components, objectsOnly=1)[0]
		ctype = cls.getComponentType(components)
		try:
			A, B = components = cls.convertComponentType(components, ctype)[:2]
		except ValueError as error:
			print ('File "{}" in getShortestPath\n# Error: Operation requires exactly two components. #\n	{}'.format(__file__, error))
			return []

		returnAsVerts=False
		if ctype=='vtx':
			edgesA = cls.convertComponentType(A, 'e', flatten=1)
			vertsA = cls.convertComponentType(edgesA, 'vtx', flatten=1)
			closestA = cls.getClosestVerts(B, [i for i in vertsA if not i==A])[0]
			edgeA = [e for e in edgesA if closestA[1] in cls.convertComponentType(e, 'vtx', flatten=1)]

			edgeB = cls.convertComponentType(B, 'e', flatten=1)
			vertsB = cls.convertComponentType(edgeB, 'vtx', flatten=1)
			closestB = cls.getClosestVerts(A, [i for i in vertsB if not i==B])[0]
			edgeB = [e for e in edgeB if closestB[1] in cls.convertComponentType(e, 'vtx', flatten=1)]

			components = (edgeA, edgeB)
			ctype = 'e'
			returnAsVerts = True

		compNums = cls.convertComponentType(components, ctype, returnType='int', flatten=True)

		kwargs = {
			'shortestFacePath' if ctype=='f' 
			else 'shortestEdgePathUV' if ctype=='uv' 
			else 'shortestEdgePath': compNums
		}
		compLong = set(pm.polySelect(obj, query=1, **kwargs) + compNums)

		result = cls.convertIntToComponent(obj, compLong, ctype, returnType=returnType, flatten=flatten)

		if returnAsVerts:
			result = cls.convertComponentType(result, 'vtx', flatten=flatten)

		return result


	@classmethod
	def getEdgesByNormalAngle(cls, objects, lowAngle=50, highAngle=130, returnType='str', flatten=False):
		'''Get a list of edges having normals between the given high and low angles using maya's polySelectConstraint.

		:Parameters:
			objects (str)(list)(obj) = The object(s) to get edges of.
			lowAngle (int) = Normal angle low range.
			highAngle (int) = Normal angle high range.
			returnType (str) = The desired returned object type. 
				valid: 'str'(default), 'obj', 'int'(valid only at sub-object level)
			flatten (bool) = Flattens the returned list of objects so that each component is it's own element.

		:Return:
			(list) Polygon edges.

		ex. call: getEdgesByNormalAngle('cyl', 50, 130) #returns: ['cylShape.e[0:23]']
		'''
		orig_selection = pm.ls(sl=1) #get currently selected objects in order to re-select them after the contraint operation.

		pm.select(objects)
		pm.polySelectConstraint(angle=True, anglebound=(lowAngle, highAngle), mode=3, type=0x8000) #Constrain that selection to only edges of a certain Angle
		pm.selectType(polymeshEdge=True)

		mask = cls.convertAlias('edges', 'int')
		objects = pm.filterExpand(selectionMask=mask, expand=flatten)

		edges = cls.getComponents(objects, componentType='edges', returnType=returnType, flatten=flatten) #get selected edges with constraint active.
		pm.polySelectConstraint(mode=0) #Remove the selection constraint.

		pm.select(orig_selection) #re-select any originally selected objects.
		return edges


	@classmethod
	def getComponentsByNumberOfConnected(cls, components, num_of_connected=(0,2), connectedType='', returnType='str'):
		'''Get a list of components filtered by the number of their connected components.

		:Parameters:
			components (str)(list)(obj) = The components to filter.
			num_of_connected (int)(tuple) = The number of connected components. Can be given as a range. (Default: (0,2))
			connectedType (str)(int) = The desired component mask. (valid: 'vtx','vertex','vertices','Polygon Vertex',31,0x0001(vertices), 'e','edge','edges','Polygon Edge',32,0x8000(edges), 'f','face','faces','Polygon Face',34,0x0008(faces), 'uv','texture','texture coordinates','Polygon UV',35,0x0010(texture coordiantes).
			returnType (str) = The desired returned object type. 
				valid: 'str'(default), 'obj', 'int'(valid only at sub-object level)

		:Return:
			(list) flattened list.

		ex. faces = getComponentsByNumberOfConnected('sph.f[:]', 4, 'e') #returns faces with four connected edges (four sided faces).
		ex. verts = getComponentsByNumberOfConnected('pln.vtx[:]', (0,2), 'e') #returns vertices with up to two connected edges.
		'''
		try:
			lowRange, highRange = num_of_connected
		except TypeError as error:
			lowRange = highRange = num_of_connected

		typ = cls.getComponentType(components)
		if connectedType:
			ctype = cls.convertAlias(connectedType)
		else:
			ctype = typ

		result=[]
		for c in pm.ls(components, flatten=True):

			attached = cls.convertComponentType(c, ctype, flatten=True)
			n = len(attached)
			if n>=lowRange and n<=highRange:
				result.append(c)

		result = convertArrayType(result, returnType=returnType)
		return result


	@classmethod
	def getVertexNormal(cls, vertex, angleWeighted=False):
		'''Return the normal at the given vertex. The returned normal is a single 
		per-vertex normal, so unshared normals at a vertex will be averaged.

		:Parameters:
			vertex (str)(obj)(list) = A polygon vertex.
			angleWeighted (bool) = Weight by the angle subtended by the face at the vertex. 
				If angleWeighted is set to false, a simple average of surround face normals is returned.
				The simple average evaluation is significantly faster than the angle-weighted average.
		:Return:
			(MVector)
		'''
		import maya.api.OpenMaya as om

		mesh = pm.ls(vertex, objectsOnly=True)[0].name()
		selectionList = om.MSelectionList() #empty selection list.
		selectionList.add(mesh)

		dagPath = selectionList.getDagPath(0) #create empty dag path object.
		mesh = om.MFnMesh(dagPath) #get mesh.

		vtxID = convertArrayType(vertex, 'int')[0]
		return mesh.getVertexNormal(vtxID, angleWeighted, space=om.MSpace.kWorld) #get vertex normal and use om.MSpace.kObject for object space.


	@staticmethod
	def getVectorFromComponents(components):
		'''Get a vector representing the averaged and normalized vertex-face normals.

		:Parameters:
			components (list) = A list of component to get normals of.

		:Return:
			(tuple) vector ie. (-4.5296159711938344e-08, 1.0, 1.6846732009412335e-08)
		'''
		vertices = pm.polyListComponentConversion(components, toVertex=1)

		norm = pm.polyNormalPerVertex(vertices, query=True, xyz=True) #return all of the normals associated with the vert.
		normal_vector = (sum(norm[0::3])/len(norm[0::3]), sum(norm[1::3])/len(norm[1::3]), sum(norm[2::3])/len(norm[2::3])) #averaging of all x,y,z points.

		return normal_vector

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Return:
		(obj)
	'''
	try:
		return getattr(Cmpt, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')







# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------


# def filterComponents(cls, frm, inc=[], exc=[]):
# 		'''Filter the given 'frm' list for the items in 'exc'.

# 		:Parameters:
# 			frm (str)(obj)(list) = The components(s) to filter.
# 			inc (str)(obj)(list) = The component(s) to include.
# 			exc (str)(obj)(list) = The component(s) to exclude.
# 								(exlude take precidence over include)
# 		:Return:
# 			(list)

# 		ex. call: filterComponents('obj.vtx[:]', 'obj.vtx[1:23]') #returns: [MeshVertex('objShape.vtx[0]'), MeshVertex('objShape.vtx[24]'), MeshVertex('objShape.vtx[25]')]
# 		'''
# 		exc = pm.ls(exc, flatten=True)
# 		if not exc:
# 			return frm

# 		c, *other = components = pm.ls(frm, flatten=True)
# 		#determine the type of items in 'exc' by sampling the first element.
# 		if isinstance(c, str):
# 			if 'Shape' in c:
# 				rtn = 'transform'
# 			else:
# 				rtn = 'str'
# 		elif isinstance(c, int):
# 			rtn = 'int'
# 		else:
# 			rtn = 'obj'

# 		if exc and isinstance(exc[0], int): #attempt to create a component list from the given integers. warning: this will only exclude from a single object.
# 			obj = pm.ls(frm, objectsOnly=1)
# 			if len(obj)>1:
# 				return frm
# 			componentType = cls.getComponentType(frm[0])
# 			typ = cls.convertAlias(componentType) #get the correct componentType variable from possible args.
# 			exc = ["{}.{}[{}]".format(obj[0], typ, n) for n in exc]

# 		if inc and isinstance(inc[0], int): #attempt to create a component list from the given integers. warning: this will only exclude from a single object.
# 			obj = pm.ls(frm, objectsOnly=1)
# 			if len(obj)>1:
# 				return frm
# 			componentType = cls.getComponentType(frm[0])
# 			typ = cls.convertAlias(componentType) #get the correct componentType variable from possible args.
# 			inc = ["{}.{}[{}]".format(obj[0], typ, n) for n in inc]

# 		inc = convertArrayType(inc, returnType=rtn, flatten=True) #assure both lists are of the same type for comparison.
# 		exc = convertArrayType(exc, returnType=rtn, flatten=True)
# 		return [i for i in components if i not in exc and (inc and i in inc)]


# !/usr/bin/python
# coding=utf-8
import os, sys
import unittest
import inspect

import pymel.core as pm

from mayatk import *
import test


sfr = pm.melGlobals['cmdScrollFieldReporter']
pm.cmdScrollFieldReporter(sfr, edit=1, clear=1)


class Main(unittest.TestCase):
	'''
	'''
	def perform_test(self, case):
		'''
		'''
		for expression, expected_result in case.items():
			m = str(expression).split('(')[0] #ie. 'self.setCase' from "self.setCase('xxx', 'upper')"

			try:
				path = os.path.abspath(inspect.getfile(eval(m)))
			except TypeError as error:
				path = ''

			result = eval(expression)
			self.assertEqual(
				result, 
				expected_result, 
				"\n\nError: {}\n  Call:     {}\n  Expected: {} {}\n  Returned: {} {}".format(path, expression.replace('self.', '', 1), type(expected_result), expected_result, type(result), result)
			)


class Rigtk_test(Main, Rigtk):
	'''
	'''
	#Tear down the any previous test by creating a new scene:
	pm.mel.file(new=True, force=True)

	#assemble the test scene:
	if not pm.objExists('loc'):
		loc = pm.spaceLocator(name='loc')

	if not pm.objExists('cyl'):
		cyl = pm.polyCylinder(radius=5, height=10, subdivisionsX=6, subdivisionsY=1, subdivisionsZ=1, name='cyl')

	def test_createLocator(self):
		'''
		'''
		self.perform_test({
			"self.createLocator('_loc')": '_loc',
		})

	def test_removeLocator(self):
		'''
		'''
		self.perform_test({
			"self.removeLocator('loc')": None,
		})

	def test_resetPivotTransforms(self):
		'''
		'''
		self.perform_test({
			"self.resetPivotTransforms('cyl')": None,
		})

	def test_bakeCustomPivot(self):
		'''
		'''
		self.perform_test({
			"self.bakeCustomPivot('cyl')": None,
			"self.bakeCustomPivot('cyl', position=True)": None,
			"self.bakeCustomPivot('cyl', orientation=True)": None,
		})

	def test_setAttrLockState(self):
		'''
		'''
		self.perform_test({
			"self.setAttrLockState('cyl')": None,
		})

	def test_createGroup(self):
		'''
		'''
		self.perform_test({
			"self.createGroup(name='emptyGrp').name()": 'emptyGrp',
		})

	def test_createGroupLRA(self):
		'''
		'''
		self.perform_test({
			"self.createGroupLRA('cyl', 'LRAgrp').name()": 'LRAgrp',
		})

	def test_createLocatorAtObject(self):
		'''
		'''
		self.perform_test({
			"self.createLocatorAtObject('cyl')": None,
		})

# -----------------------------------------------------------------------------

if __name__=='__main__':

	unittest.main(exit=False)






# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

# """
# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})


# def test_(self):
# 	'''
# 	'''
# 	self.perform_test({
# 		# "self.": '',
# 	})
# """

# # Deprecated ---------------------